#!/usr/bin/env python
"""This module has tests for the vis.pds.labelmaker.collection functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from datetime import datetime, timezone
from pathlib import Path
import unittest
from unittest.mock import create_autospec, patch

import vipersci.pds.labelmaker.collection as co
from vipersci import util


class TestCollection(unittest.TestCase):
    def setUp(self):
        self.base = {
            "collection_lid": "urn.pds:collection",
            "investigation_name": "nameo",
            "investigation_type": "typeo",
            "investigation_lid": "urn:pds:nameo",
            "host_name": "hosto",
            "host_lid": "urn:pds:hosto",
            "target_name": "Moon",
            "target_type": "Satellite",
            "target_lid": "urn:pds:Moon",
        }

        self.label1 = {
            "vid": "0.1",
            "instruments": {"urn:pds:leftcam": "NavCam Left"},
            "purposes": [
                "Science",
            ],
            "processing_levels": [
                "Raw",
            ],
            "start_date_time": datetime(2023, 10, 1, 20, 00, 00, tzinfo=timezone.utc),
            "stop_date_time": datetime(2023, 10, 1, 20, 00, 1, tzinfo=timezone.utc),
        }
        self.label1.update(self.base)

        self.label2 = {
            "vid": "2.0",
            "instruments": {"urn:pds:rightcam": "NavCam Right"},
            "purposes": [
                "Science",
            ],
            "processing_levels": [
                "Derived",
            ],
            "start_date_time": datetime(2023, 11, 1, 20, 00, 00, tzinfo=timezone.utc),
            "stop_date_time": datetime(2023, 11, 1, 20, 00, 1, tzinfo=timezone.utc),
        }
        self.label2.update(self.base)

    def test_add_parser(self):
        parser = ArgumentParser()
        subparsers = parser.add_subparsers()
        co.add_parser(subparsers)

        d = vars(
            parser.parse_args(
                [
                    "collection",
                    "--config",
                    "dumb.yml",
                    "-t",
                    "template.xml",
                    "file1.xml",
                    "file2.xml",
                ]
            )
        )
        self.assertIn("config", d)
        self.assertIn("template", d)
        self.assertIn("labelfiles", d)

    @patch("vipersci.pds.labelmaker.collection.ET.fromstring")
    @patch(
        "vipersci.pds.labelmaker.collection.get_common_label_info",
        return_value={"lid": "urn:pds:dummy:pid"},
    )
    def test_get_label_info(self, m_gcli, m_fromstring):
        p = create_autospec(Path)
        d = co.get_label_info(p)

        m_fromstring.assert_called_once()
        m_gcli.assert_called_once()
        self.assertEqual(d["collection_lid"], "urn:pds:dummy")

    def test_check_and_derive(self):
        config = self.base.copy()
        config["collection_type"] = "Data"
        config["modification_details"] = [{"version": "1.0"}, {"version": "2.0"}]

        d = co.check_and_derive(config, [self.label1, self.label2])

        self.assertEqual(
            {
                "vid": "2.0",
                "instruments": {
                    "urn:pds:leftcam": "NavCam Left",
                    "urn:pds:rightcam": "NavCam Right",
                },
                "purposes": {"Science"},
                "processing_levels": {"Derived", "Raw"},
                "start_date_time": "2023-10-01T20:00:00Z",
                "stop_date_time": "2023-11-01T20:00:01Z",
            },
            d,
        )

    def test_main(self):
        parser = ArgumentParser(parents=[util.parent_parser()])
        subparsers = parser.add_subparsers()
        co.add_parser(subparsers)

        args = parser.parse_args(
            [
                "collection",
                "-c",
                "dummy.yml",
                "-t",
                "template.xml",
                "file1.xml",
                "file2.xml",
            ]
        )

        args.config = create_autospec(Path)

        config = self.base.copy()
        config["collection_type"] = "Data"
        config["modification_details"] = [{"version": "1.0"}, {"version": "2.0"}]

        path_mock = create_autospec(Path)
        path_mock.exists.return_value = False

        with patch(
            "vipersci.pds.labelmaker.collection.yaml.safe_load", return_value=config
        ) as m_yamlload, patch(
            "vipersci.pds.labelmaker.collection.get_label_info",
            side_effect=(self.label1, self.label2),
        ) as m_get_label, patch(
            "vipersci.pds.labelmaker.collection.write_inventory"
        ) as m_write_inventory, patch(
            "vipersci.pds.labelmaker.collection.write_xml"
        ) as m_write_xml, patch(
            "vipersci.pds.labelmaker.collection.Path", return_value=path_mock
        ) as m_path:
            co.main(args)

            m_yamlload.assert_called_once()
            self.assertEqual(m_get_label.call_count, 2)
            self.assertEqual(m_path.call_count, 2)
            m_write_inventory.assert_called_once()
            m_write_xml.assert_called_once()

        # The collection XML label already exists:
        path_mock.exists.return_value = True
        with patch(
            "vipersci.pds.labelmaker.collection.yaml.safe_load", return_value=config
        ), patch("vipersci.pds.labelmaker.collection.Path", return_value=path_mock):
            self.assertRaises(FileExistsError, co.main, args)

        # The collection XML label doesn't exist, but the collection inventory does.
        path_mock2 = create_autospec(Path)
        path_mock2.exists.side_effect = (False, True)
        with patch(
            "vipersci.pds.labelmaker.collection.yaml.safe_load", return_value=config
        ), patch(
            "vipersci.pds.labelmaker.collection.get_label_info",
            side_effect=(self.label1, self.label2),
        ), patch(
            "vipersci.pds.labelmaker.collection.Path", return_value=path_mock2
        ):
            self.assertRaises(FileExistsError, co.main, args)



#!/usr/bin/env python
"""This module has tests for the vipersci.pds.bundle_install functions."""

# Copyright 2022-2024, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
from textwrap import dedent
import unittest
from unittest.mock import create_autospec, mock_open, patch
import xml.etree.ElementTree as ET

import vipersci.pds.bundle_install as bi


class TestCLI(unittest.TestCase):
    def test_arg_parser(self):
        p = bi.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        self.assertRaises(SystemExit, p.parse_args)
        d = vars(p.parse_args(["src_dir", "bld_dir"]))
        self.assertIn("source_directory", d)
        self.assertIn("build_directory", d)

    def test_main(self):
        pa_ret_val = bi.arg_parser().parse_args(["src_dir", "build_dir"])
        bundle = ET.fromstring(
            dedent(
                """\
                <?xml version="1.0" encoding="UTF-8"?>
                <?xml-model href="http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1L00.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?><Product_Bundle xmlns="http://pds.nasa.gov/pds4/pds/v1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://pds.nasa.gov/pds4/pds/v1 http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1L00.xsd">
                <Identification_Area>
                    <logical_identifier>urn:nasa:pds:viper_vis</logical_identifier>
                    <version_id>99.99</version_id>
                    <title>VIPER Visible Imaging System Products Bundle</title>
                    <information_model_version>1.21.0.0</information_model_version>
                    <product_class>Product_Bundle</product_class>
                    <Citation_Information>
                        <author_list>Ross Beyer and Uland Wong</author_list>
                        <publication_year>2023</publication_year>
                        <description>VIPER Visible Imaging System Products Bundle</description>
                    </Citation_Information>
                    <Modification_History>
                        <Modification_Detail>
                            <modification_date>2023-10-28</modification_date>
                            <version_id>99.99</version_id>
                            <description>Bogus testing version</description>
                        </Modification_Detail>
                    </Modification_History>
                </Identification_Area>
                <Context_Area>
                    <Time_Coordinates>
                        <start_date_time>2023-10-26T20:00:00Z</start_date_time>
                        <stop_date_time>2023-11-09T17:05:00.000511Z</stop_date_time>
                    </Time_Coordinates>
                    <Primary_Result_Summary>
                        <purpose>Science</purpose>
                        <processing_level>Raw</processing_level><processing_level>Derived</processing_level>
                    </Primary_Result_Summary>
                    <Investigation_Area>
                        <name>VIPER</name>
                        <type>Mission</type>
                        <Internal_Reference>
                            <lid_reference>urn:nasa:pds:viper</lid_reference>
                            <reference_type>bundle_to_investigation</reference_type>
                        </Internal_Reference>
                    </Investigation_Area>
                    <Observing_System>
                        <Observing_System_Component>
                            <name>VIPER</name>
                            <type>Host</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.viper</lid_reference>
                                <reference_type>is_instrument_host</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component>
                        <Observing_System_Component>
                            <name>NavCam Left</name>
                            <type>Instrument</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.viper.navcam_left</lid_reference>
                                <reference_type>is_instrument</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component><Observing_System_Component>
                            <name>NavCam Right</name>
                            <type>Instrument</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument:viper.vis</lid_reference>
                                <reference_type>is_instrument</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component>
                    </Observing_System>
                    <Target_Identification>
                        <name>Moon</name>
                        <type>Satellite</type>
                        <Internal_Reference>
                            <lid_reference>urn:nasa:pds:context:target:satellite.earth.moon</lid_reference>
                            <reference_type>bundle_to_target</reference_type>
                        </Internal_Reference>
                    </Target_Identification>
                </Context_Area>
                <Bundle>
                    <bundle_type>Archive</bundle_type>
                </Bundle>
                <Bundle_Member_Entry>
                    <lidvid_reference>urn:nasa:pds:viper_vis:data_raw::99.99</lidvid_reference>
                    <member_status>Primary</member_status>
                    <reference_type>bundle_has_data_collection</reference_type>
                </Bundle_Member_Entry>
                <File_Area_Text>
                    <File>
                        <file_name>readme</file_name>
                    </File>
                </File_Area_Text>
                </Product_Bundle>
            """  # noqa: E501
            )
        )
        collection = ET.fromstring(
            dedent(
                """\
                <?xml version="1.0" encoding="UTF-8"?>
                <?xml-model href="http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1L00.sch" schematypens="http://purl.oclc.org/dsdl/schematron"?><Product_Collection xmlns="http://pds.nasa.gov/pds4/pds/v1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://pds.nasa.gov/pds4/pds/v1 http://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1L00.xsd">
                <Identification_Area>
                        <logical_identifier>urn:nasa:pds:viper_vis:data_raw</logical_identifier>
                        <version_id>99.99</version_id>
                        <title>VIPER Visible Imaging System Raw Data Collection</title>
                    <information_model_version>1.21.0.0</information_model_version>
                    <product_class>Product_Collection</product_class>
                    <Citation_Information>
                        <author_list>Ross Beyer and Uland Wong</author_list>
                        <publication_year>2023</publication_year>
                        <description>VIPER Visible Imaging System Raw Data Collection</description>
                    </Citation_Information>
                    <Modification_History>
                        <Modification_Detail>
                            <modification_date>2023-10-28</modification_date>
                            <version_id>99.99</version_id>
                            <description>Bogus testing version</description>
                        </Modification_Detail>
                    </Modification_History>
                </Identification_Area>
                <Context_Area>
                    <Time_Coordinates>
                        <start_date_time>2023-10-26T20:00:00Z</start_date_time>
                        <stop_date_time>2023-11-09T17:05:00.000511Z</stop_date_time>
                    </Time_Coordinates>
                    <Primary_Result_Summary>
                        <purpose>Science</purpose>
                        <processing_level>Raw</processing_level>
                    </Primary_Result_Summary>
                    <Investigation_Area>
                        <name>VIPER</name>
                        <type>Mission</type>
                        <Internal_Reference>
                            <lid_reference>urn:nasa:pds:viper</lid_reference>
                            <reference_type>collection_to_investigation</reference_type>
                        </Internal_Reference>
                    </Investigation_Area>
                    <Observing_System>
                      <Observing_System_Component>
                            <name>VIPER</name>
                            <type>Host</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.viper</lid_reference>
                                <reference_type>is_instrument_host</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component>
                        <Observing_System_Component>
                            <name>NavCam Left</name>
                            <type>Instrument</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument:viper.vis</lid_reference>
                                <reference_type>is_instrument</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component><Observing_System_Component>
                            <name>NavCam Right</name>
                            <type>Instrument</type>
                            <Internal_Reference>
                                <lid_reference>urn:nasa:pds:context:instrument:viper.vis</lid_reference>
                                <reference_type>is_instrument</reference_type>
                            </Internal_Reference>
                        </Observing_System_Component>
                    </Observing_System>
                    <Target_Identification>
                        <name>Moon</name>
                        <type>Satellite</type>
                        <Internal_Reference>
                            <lid_reference>urn:nasa:pds:context:target:satellite.earth.moon</lid_reference>
                            <reference_type>collection_to_target</reference_type>
                        </Internal_Reference>
                    </Target_Identification>
                </Context_Area>
                <Collection>
                    <collection_type>Data</collection_type>
                </Collection>
                <File_Area_Inventory>
                    <File>
                        <file_name>collection_data_raw.csv</file_name>
                        <creation_date_time>2024-01-29T22:07:08.297641Z</creation_date_time>
                    </File>
                    <Inventory>
                        <offset unit="byte">0</offset>
                        <parsing_standard_id>PDS DSV 1</parsing_standard_id>
                        <records>10</records>
                        <record_delimiter>Carriage-Return Line-Feed</record_delimiter>
                        <field_delimiter>Comma</field_delimiter>
                        <Record_Delimited>
                            <fields>2</fields>
                            <groups>0</groups>
                            <Field_Delimited>
                                <name>Member Status</name>
                                <field_number>1</field_number>
                                <data_type>ASCII_String</data_type>
                                <maximum_field_length unit="byte">1</maximum_field_length>
                                <description>
                                  P indicates primary member of the collection
                                  S indicates secondary member of the collection
                                </description>
                            </Field_Delimited>
                            <Field_Delimited>
                                <name>LIDVID_LID</name>
                                <field_number>2</field_number>
                                <data_type>ASCII_LIDVID_LID</data_type>
                                <maximum_field_length unit="byte">255</maximum_field_length>
                                <description>
                                  The LID or LIDVID of a product that is a member
                                  of the collection.
                                </description>
                            </Field_Delimited>
                        </Record_Delimited>
                        <reference_type>inventory_has_member_product</reference_type>
                    </Inventory>
                </File_Area_Inventory>
                </Product_Collection>
                """  # noqa: E501
            )
        )
        inventory = dedent(
            """\
        P,urn:nasa:pds:viper_vis:data_raw:240225-200000-ncl-d::99.99
        """
        )
        with patch("vipersci.pds.bundle_install.arg_parser") as parser, patch(
            "vipersci.pds.bundle_install.copy2"
        ), patch(
            "vipersci.pds.bundle_install.ET.fromstring",
            side_effect=[bundle, collection],
        ), patch(
            "vipersci.pds.bundle_install.open", mock_open(read_data=inventory)
        ), patch(
            "vipersci.pds.bundle_install.get_lidvidfile",
            return_value={
                "lid": "urn:nasa:pds:viper_vis:data_raw:240225-200000-ncl-d",
                "vid": "99.99",
                "productfile": "240225-200000-ncl-d.tif",
            },
        ):
            build_path = create_autospec(Path)
            src_path = create_autospec(Path)
            src_col_dir = create_autospec(Path)
            src_col_dir.rglob.return_value = [
                create_autospec(Path),
            ]
            src_path.__truediv__.return_value = src_col_dir
            pa_ret_val.build_directory = build_path
            pa_ret_val.source_directory = src_path
            parser.return_value.parse_args.return_value = pa_ret_val
            bi.main()



"""Various helper functions for VIS PDS operations.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import date
from importlib import resources
import logging
from pathlib import Path

from genshi.template import MarkupTemplate

logger = logging.getLogger(__name__)

lids = {
    "bundle": "urn:nasa:pds:viper_vis",
    "mission": "urn:nasa:pds:viper",
    "spacecraft": "urn:nasa:pds:context:instrument_host:spacecraft.viper",
    "instrument": "urn:nasa:pds:context:instrument:viper.vis",
}


def version_info():
    # This should reach into a database and do something smart to figure
    # out how to populate this, but for now, hardcoding:
    d = {
        "modification_details": [
            {
                "version": 99.99,
                "date": date.today().isoformat(),
                "description": "Dumb information model forces this to be a 'valid' "
                "version instead of 0.1 for testing, which should be "
                "allowed.",
            }
        ],
        "vid": 99.99,
    }
    return d


def write_xml(
    product: dict,
    template: str,
    outdir: Path = Path.cwd(),
):
    """
    Writes a PDS4 XML label in *outdir* based on the contents of
    the *product* object, which must be of type Raw_Product.

    The *template_path* can be a path to an appropriate template
    XML file, but defaults to the raw-template.xml file provided
    with this library.
    """
    if Path(template).exists():
        tmpl = MarkupTemplate(Path(template).read_text())
    else:
        tmpl = MarkupTemplate(
            resources.read_text("vipersci.vis.pds.data", str(template))
        )

    d = version_info()
    d.update(product)

    logger.info(d)

    stream = tmpl.generate(**d)
    out_path = (outdir / product["product_id"]).with_suffix(".xml")
    out_path.write_text(stream.render())
    return



# -*- coding: utf-8 -*-
"""
Takes a traverse of positions and the following maps: Insolation, Surface
Temperatures, and Ice Stability Depth, and simulates
what the OH and H2O band depths would be.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
import logging
from pathlib import Path

import numpy as np
import rasterio

from vipersci import nirvss
from vipersci.carto.nss_modeler import write_tif

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-b",
        "--burial_depth",
        type=Path,
        required=True,
        help="A GeoTIFF map whose pixels are the burial depth of the water layer.",
    )
    parser.add_argument(
        "-i",
        "--insolation",
        type=Path,
        required=True,
        help="A GeoTIFF map whose pixels are the insolation value of the surface.",
    )
    parser.add_argument(
        "--temperature",
        type=Path,
        required=True,
        help="A GeoTIFF map whose pixels are the surface temperature.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="The output CSV file or if -t is not given, this will be used "
        "as the prefix for the two output maps, which will end in "
        "_d1.tif and _d2.tif .",
    )
    parser.add_argument("-t", "--traverse", type=Path, help="Input CSV file.")

    return parser


class LocationSimulator:
    def __init__(
        self,
        bd_map: Path,
        insl_map: Path,
        temp_map: Path,
    ):
        """
        :param bd_map: The path to the Burial Depth map.
        :type bd_map: Path
        :param insl_map: The path to the Insolation map.
        :type insl_map: Path
        :param temp_map: The path to the Surface Temperature map.
        :type temp_map: Path

        """
        self.bd_aff, self.bd_arr = self._init_map(bd_map)
        self.insl_aff, self.insl_arr = self._init_map(insl_map)
        self.temp_aff, self.temp_arr = self._init_map(temp_map)

        return

    @staticmethod
    def _init_map(raster: Path):
        # Helper function for __init__()
        data = rasterio.open(raster)
        return data.transform, data.read(1)

    def __call__(self, xycoords: np.ndarray, poisson=False):
        """
        Returns simulated H2O and OH band depth values at the given
        location(s).

        :param xycoords:  A sequence or np.array object of shape (2, ...)
        indicating the projected x, y coordinate to simulate the data for.
        :returns: A two-tuple of values or np.arrays.  If xycoords contained
        only two elements, a two-tuple is returned of the H2O and
        OH band depth values at that location.  If xycoords contains more than
        one  location, then the returned two-tuple will be a numpy array
        of H20 band depth values, and a numpy array of OH band depth values.
        """
        bd_rows, bd_cols = rasterio.transform.rowcol(
            self.bd_aff, xycoords[0], xycoords[1]
        )
        insl_rows, insl_cols = rasterio.transform.rowcol(
            self.insl_aff, xycoords[0], xycoords[1]
        )
        temp_rows, temp_cols = rasterio.transform.rowcol(
            self.temp_aff, xycoords[0], xycoords[1]
        )

        bd_vals = self.bd_arr[bd_rows, bd_cols]
        insl_vals = self.insl_arr[insl_rows, insl_cols]
        temp_vals = self.temp_arr[temp_rows, temp_cols]

        h2o_vals = nirvss.band_depth_H2O(temp_vals, bd_vals)
        oh_vals = nirvss.band_depth_OH(insl_vals)

        return h2o_vals, oh_vals


def main():
    args = arg_parser().parse_args()

    if args.traverse is None:
        # Make ideal detector maps
        bd_data = rasterio.open(args.burial_depth)
        insl_data = rasterio.open(args.insolation)
        temp_data = rasterio.open(args.temperature)

        for a, b in zip((bd_data, insl_data), (insl_data, temp_data)):
            for prop in ("width", "height", "transform"):
                if getattr(a, prop) != getattr(b, prop):
                    raise ValueError(
                        f"The {prop} value of the {a.name} and {b.name} maps are "
                        "different.  They must be the same to output detector maps."
                    )

        h2o = nirvss.band_depth_H2O(
            temp_data.read(1).flatten(), bd_data.read(1).flatten()
        )
        oh = nirvss.band_depth_OH(insl_data.read(1).flatten())

        h2o_arr = h2o.reshape(bd_data.shape)
        oh_arr = oh.reshape(bd_data.shape)

        kwds = bd_data.profile
        write_tif(args.output, "_h2o.tif", h2o_arr, kwds)
        write_tif(args.output, "_oh.tif", oh_arr, kwds)
        return

    else:
        # Make an output traverse file.
        simulator = LocationSimulator(
            args.burial_depth,
            args.insolation,
            args.temperature,
        )

        with open(args.traverse, newline="") as csvin, open(
            args.output, "w", newline=""
        ) as csvout:
            reader = csv.DictReader(csvin)  # expecting at least x and y cols.
            fieldnames = reader.fieldnames + [
                "bd_h2o",
                "bd_oh",
            ]
            writer = csv.DictWriter(csvout, fieldnames=fieldnames)
            writer.writeheader()

            # Would it be faster to read in all the points, feed them all to the
            # simulator, and then write out points?  Maybe, iterating for now:
            for row in reader:
                coords = np.array((row["x"], row["y"]), dtype=float)
                # print(coords)
                # print(coords.shape)
                row["bd_h2o"], row["bd_oh"] = simulator(coords)

                writer.writerow(row)

    return



#!/usr/bin/env python
"""This module has tests for the light_records module."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest

from datetime import datetime, timezone

from vipersci.vis.db import light_records as lr


class TestLights(unittest.TestCase):
    def setUp(self):
        self.dt = datetime(2023, 1, 1, 0, 0, 0, tzinfo=timezone.utc)

    def test_init(self):
        good = lr.LightRecord(
            name="navLeft",
            on=True,
            datetime=self.dt,
        )
        self.assertEqual(good.name, "NavLight Left")

    def test_init_fail(self):
        self.assertRaises(
            ValueError,
            lr.LightRecord,
            name="not a name",
            on=False,
            datetime=self.dt,
        )

        self.assertRaises(
            ValueError,
            lr.LightRecord,
            name="navRight",
            on=True,
            datetime="not a datetime",
        )



"""
This heatmaps module takes scalar values with 2D coordinates and creates a density
heatmap representation with individual points effectively averaged together.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import math
from multiprocessing import Pool
from itertools import chain
import time
from typing import Any, Dict, Optional, Sequence, Tuple
import logging

import pyproj
import numpy as np
from numpy.typing import NDArray
import rasterio
import rasterio.features
import shapely.geometry
from sklearn.neighbors import KernelDensity

from vipersci.carto.bounds import compute_bounds, pad_grid_align_bounds

logger = logging.getLogger(__name__)


def buffered_mask(
    linestring: shapely.geometry.LineString,
    transform: rasterio.Affine,
    buffer: float,
    all_touched=False,
) -> NDArray[np.bool_]:
    """
    Returns a boolean numpy array from rasterio.features.geometry_mask() that
    indicates the shape of *linestring* buffered by *buffer* interpreted into
    a grid defined by *transform*.

    In general, the returned array can be used as a numpy mask, where pixels
    that overlap the shapes are False.

    The parameter *all_touched* is passed directly to
    rasterio.features.geometry_mask(), and you can read more about it there.
    """
    logger.debug("Start buffer")
    start = time.perf_counter()
    logger.debug(linestring.bounds)
    buffered = linestring.buffer(buffer, resolution=2)
    if isinstance(buffered, shapely.geometry.Polygon):
        geometries = [
            buffered,
        ]
    else:  # MultiPolygon
        geometries = buffered.geoms
    logger.debug(f"Buffered geometries in {time.perf_counter() - start:.6f}s")

    logger.debug(buffered.bounds)

    window = rasterio.windows.from_bounds(*buffered.bounds, transform).round_lengths(
        op="ceil"
    )

    logger.debug(window)

    logger.debug("Start geometry_mask")
    start = time.perf_counter()
    mask = rasterio.features.geometry_mask(
        geometries,
        rasterio.windows.shape(window),
        transform,
        all_touched=all_touched,
        invert=False,
    )
    logger.debug(f"Created geometry_mask in {time.perf_counter() - start:.6f}s")

    return mask


def as_ndarray(input: Sequence) -> NDArray:
    """
    Check the type of the input and return it as an np ndarray, converting if needed
    Parameters:
        input
    Returns:
        ndarray
    """
    if isinstance(input, np.ndarray):
        return input
    else:
        return np.asarray(input)


def generate_density_heatmap(
    x_coords: Sequence,  # list or np.ndarray
    y_coords: Sequence,  # list or np.ndarray
    values: Sequence,  # list or np.ndarray
    gsd: float = 1,  # in same units of x, y
    radius: float = 1,  # radius of "data disk", in units of x, y
    padding: Optional[float] = None,  # in pixels
    nodata_value: float = 0,
    transform: Optional[rasterio.Affine] = None,
    processes: int = 1,
    sample_bounds: Optional[shapely.geometry.Polygon] = None,
    frequencies: Optional[NDArray[np.float32]] = None,
) -> Tuple[
    rasterio.Affine,
    NDArray[np.uintc],
    NDArray[np.float32],
    NDArray[np.float32],
]:
    """
    Perform tophat kernel density estimation to build a continuous heatmap
    representation of scalar values with 2d coordinates.

    Parameters:
        x_coords: x coordinates of the data points
        y_coords: y coordinates of the data points
        values: values of the data points.  Any None or np.nan values will be ignored
            in the kernel density estimation.
        gsd: Coordinate interval at which to sample the output distribution,
            or the ground sample distance of the output arrays.
            Defaults to 1.
        radius:  This is the kernel bandwidth used in the density estimation,
            and should be the sensing "radius" of the instrument.
            Defaults to 1.
        padding: Square padding in pixels to add to the bounds of data when
            returning an array.  If None (the default), the value of *radius*
            converted to pixels will be used.
        nodata_value: Defaults to zero, but a different value can be specified.
        transform: If a rasterio Affine transform is not supplied (the
            default), then one will be generated, and returned.
        processes: Number of processes to use when sampling the output distribution.
            Defaults to None, allowing multiprocessing to decide how many to use.
        sample_bounds: polygon bounding the region in which to actually sample data.
            Defaults to None, sampling over the entire region.
        frequencies: frequency values returned from a previous call to
            generate_density_heatmap. Used when source has multiple data values
            (ie, multiple detectors) sampled at the same rate and time.
            Defaults to None, so frequencies will be computed.

    Returns:
        A tuple (transform, counts, avg, frequencies)
        transform: transform used to georeference the output data
        counts: The number of observations (data points defined by the x/y
            center location and *radius*) that overlap a given grid cell.
        avg: Averaged value of all the individual observations that overlap
            a given grid cell.
        frequencies: Can be provided to subsequent generate_density_heatmap calls
            on new data sampled at the same locations

    The *counts* and *avg* objects are numpy arrays.
    """

    if processes < 1:
        raise ValueError("Processes must be a positive integer.")

    if not (len(x_coords) == len(y_coords) == len(values)):
        raise ValueError("Input arrays must be of the same length.")

    values_all = as_ndarray(values)

    missing_idx = np.isnan(values_all.astype(float))

    x_coords_np = np.delete(as_ndarray(x_coords), np.argwhere(missing_idx))
    y_coords_np = np.delete(as_ndarray(y_coords), np.argwhere(missing_idx))
    values_np = np.delete(values_all, np.argwhere(missing_idx))

    points = shapely.geometry.LineString(np.stack((x_coords_np, y_coords_np), axis=1))
    if sample_bounds is not None:
        points = points.intersection(sample_bounds)

    if padding is None:
        buffer = radius
    else:
        # Convert from pixels of padding to a buffer in x/y distance units.
        buffer = (padding * gsd) + radius
    logger.debug(f"buffer: {buffer}")

    if transform is None:
        if sample_bounds is None:
            bounds = compute_bounds(x_coords_np, y_coords_np)
        else:
            bounds = sample_bounds.bounds
        left, _, _, top = pad_grid_align_bounds(*bounds, gsd, math.ceil(buffer / gsd))
        transform = rasterio.transform.from_origin(left, top, gsd, gsd)
    else:
        if transform.a != gsd or transform.e != gsd:
            raise ValueError(
                f"The scale factors of the transform ({transform.a}, "
                f"{transform.e}) must both equal the ground sample distance, "
                f"({gsd})."
            )

    start = time.perf_counter()
    mask = buffered_mask(points, transform, buffer)
    end = time.perf_counter()
    logger.debug(f"Created mask in {end - start:.6f}s")

    train = np.column_stack([y_coords_np, x_coords_np])
    kde = KernelDensity(
        bandwidth=radius, metric="euclidean", kernel="tophat", algorithm="auto"
    )
    start = time.perf_counter()
    kde.fit(train)
    end = time.perf_counter()
    logger.debug(f"Trained unweighted KDE in {end - start:.6f}s")

    # Now get unmasked coordinates:
    start = time.perf_counter()
    row_coords, col_coords = np.meshgrid(
        range(mask.shape[0]), range(mask.shape[1]), indexing="ij"
    )
    row_masked: np.ma.MaskedArray = np.ma.MaskedArray(row_coords, mask)
    col_masked: np.ma.MaskedArray = np.ma.MaskedArray(col_coords, mask)

    x_tosample, y_tosample = rasterio.transform.xy(
        transform, row_masked.compressed(), col_masked.compressed()
    )

    sample_coords = np.column_stack([y_tosample, x_tosample])
    end = time.perf_counter()
    logger.debug(f"Created unmasked coordinates {end - start:.6f}s")

    if frequencies is None:
        start = time.perf_counter()
        with Pool(processes=processes) as pool:
            results = pool.imap(
                kde.score_samples,
                np.array_split(sample_coords, processes),
            )
            samples = np.fromiter(
                chain.from_iterable(results),
                dtype=float,
                count=len(sample_coords),
            )
        out_unweighted = np.exp(samples)
        total_observations = len(train)
        frequencies = out_unweighted * total_observations
        frequencies = (frequencies > 1e-9) * frequencies
        end = time.perf_counter()
        logger.info(f"Sampled {samples.shape[0]} points in {end - start:.6f}s.")

    start = time.perf_counter()
    kde.fit(train, sample_weight=values_np)
    end = time.perf_counter()
    logger.debug(f"Trained weighted KDE in {end - start:.6f}s")

    start = time.perf_counter()
    with Pool(processes=processes) as pool:
        results = pool.imap(
            kde.score_samples,
            np.array_split(sample_coords, processes),
        )
        weighted_samples = np.fromiter(
            chain.from_iterable(results), dtype=float, count=len(sample_coords)
        )
    end = time.perf_counter()
    logger.info(
        f"Sampled {weighted_samples.shape[0]} weighted points in {end - start:.6f}s."
    )

    # compute our required stats
    start = time.perf_counter()
    out_weighted = np.exp(weighted_samples)
    out_weighted = (out_weighted > 1e-9) * out_weighted
    total_counts = np.sum(values_np) * out_weighted

    with np.errstate(divide="ignore", invalid="ignore"):
        avg_values = np.nan_to_num(total_counts / frequencies, posinf=0, copy=False)

    # frequencies starts as the height of cylinders with radius of bandwidth,
    # but the meaningful value is their volume
    counts = np.around(frequencies * math.pi * math.pow(radius, 2))

    end = time.perf_counter()
    logger.info(f"Computed stats in {end - start:.6f}s")

    out_avg = np.full_like(mask, nodata_value, dtype=np.float32)
    out_counts = np.full_like(mask, 0, dtype=np.uintc)
    for x, y, avg, count in zip(x_tosample, y_tosample, avg_values, counts):
        r, c = rasterio.transform.rowcol(transform, x, y)
        out_avg[r, c] = avg
        out_counts[r, c] = count

    return transform, out_counts, out_avg, frequencies


def write_geotiff_rasterio(
    out_filepath,
    dest_crs,
    transform,
    *data,
    source_crs=None,
    nodata_value=0,
    profile={},
    compress="deflate",
) -> Dict[str, Any]:
    """
    Writes 2D data to a geotiff file

    Parameters
        out_filepath: Absolute filepath for writing
        dest_crs: The Rasterio CRS that applies to the data
        transform: The affine transform used for geolocating the data
        data: 2D array of data to write to the geotiff.  Multiple arrays will be written
            to separate bands in order
        source_crs: The Rasterio CRS the data was projected from, if applicable.
            Used to calculate 'extent' in the returned info.
        nodata_value (int): the nodata value to use in the raster. Defaults to 0.
        profile (dict): Additional profile data to use with rasterio.
            Will be merged / updated with basic information about the raster itself.
            Defaults to an empty dictionary (no additional data).
        compress: Compression method to use if not specified in profile.
            Any value supported by GDAL - defaults to "deflate"
    Returns
        A dictionary that mimics the information provided by gdalinfo
    """
    unified_profile: Dict[str, Any] = {
        "driver": "GTiff",
    }

    unified_profile.update(profile)

    unified_profile.update(
        {
            "height": data[0].shape[0],
            "width": data[0].shape[1],
            "count": len(data),
            "dtype": data[0].dtype,
            "crs": dest_crs,
            "transform": transform,
            "nodata": nodata_value,
        }
    )

    if "compress" not in profile:
        unified_profile.update(compress=compress)

    with rasterio.open(out_filepath, "w", **unified_profile) as raster:
        for i, d in enumerate(data, start=1):
            raster.write(d, i)
        gdalinfo = get_gdal_info_from_rasterio(raster, source_crs)

    return gdalinfo


def get_gdal_info_from_rasterio(
    input: rasterio.DatasetReader, source_crs: pyproj.crs.CRS
) -> Dict[str, Any]:
    """
    Construct a block of gdal-info style json from the metadata about a rasterio dataset
    Parameters:
        input: loaded rasterio dataset
    Returns:
        dict
    """
    geotransform = input.transform.to_gdal()
    result = {
        "size": [input.shape[0], input.shape[1]],
        "coordinateSystem": {"proj4": input.crs.to_proj4()},
        "geoTransform": geotransform,
        "resolution": {
            "xResolution": geotransform[1],
            "yResolution": geotransform[5],
        },
        "driverShortName": input.driver,
        "metadata": {
            "IMAGE_STRUCTURE": {
                "INTERLEAVE": input.interleaving.name if input.interleaving else "None",
                "COMPRESSION": input.compression.name if input.compression else "None",
            }
        },
        "files": input.files,
    }

    bands = []
    for index in range(0, input.count):
        band = {
            "band": input.indexes[index],
            "mask": {"flags": [f.name for f in input.mask_flag_enums[index]]},
            "type": input.dtypes[index],
            "block": list(input.block_shapes[index]),
            "description": str(input.descriptions[index]),
            "noDataValue": input.nodata,
            "colorInterpretation": input.colorinterp[index].name,
        }
        bands.append(band)

    result["bands"] = bands

    if source_crs is not None:
        x = [
            input.bounds.left,
            input.bounds.left,
            input.bounds.right,
            input.bounds.right,
        ]
        y = [
            input.bounds.bottom,
            input.bounds.top,
            input.bounds.top,
            input.bounds.bottom,
        ]
        lon_bnds, lat_bnds = pyproj.Transformer.from_crs(
            input.crs, source_crs
        ).transform(x, y)

        result["extent"] = {
            "type": "Polygon",
            "coordinates": [
                [
                    [lon_bnds[0], lat_bnds[0]],
                    [lon_bnds[1], lat_bnds[1]],
                    [lon_bnds[2], lat_bnds[2]],
                    [lon_bnds[3], lat_bnds[3]],
                    [lon_bnds[0], lat_bnds[0]],
                ]
            ],
        }

    result["cornerCoordinates"] = {
        "center": [
            (input.bounds.left + input.bounds.right) / 2,
            (input.bounds.top + input.bounds.bottom) / 2,
        ],
        "lowerLeft": [input.bounds.left, input.bounds.bottom],
        "upperLeft": [input.bounds.left, input.bounds.top],
        "lowerRight": [input.bounds.right, input.bounds.bottom],
        "upperRight": [input.bounds.right, input.bounds.top],
    }

    return result


def generate_area_bin_heatmap(
    x_coords: NDArray,
    y_coords: NDArray,
    values: NDArray,
    bin_size: float = 1,
) -> Tuple[rasterio.Affine, NDArray, NDArray]:
    """
    Generates a 1-band floating point geotiff heatmap by binning data into a grid and
        averaging the values

    Parameters:
        x_coords: array of x coordinates. Should be projected
        y_coords: array of y coordinates. Should be projected
        values: array of scalar values at each provided location
        crs: coordinate reference system used to project the locations to cartesian
            coordinates
        bin_size: size of square bins in meters

    Returns:
        A tuple (transform, out_counts, out_avg)
        transform: transform used to georeference the output data
        out_counts: The number of observations (data points) in a given area
        out_avg: Averaged values of measurements in a given area
    """

    def get_transform_from_coords(
        x_coords: NDArray, y_coords: NDArray, padding=0, grid_size: float = 1
    ) -> Tuple[rasterio.Affine, rasterio.windows.Window, rasterio.coords.BoundingBox]:
        """
        Given a set of coordinates and a desired grid size, build a transform
        representing a uniform grid over the bounds of the coordinates

        Parameters
            x_coords: array of x coordinates
            y_coords: array of y coordinates
            padding: square padding value to use around the bounding box of the provided
                coordinates
            grid_size: resolution of the grid
        """
        x_min = grid_size * math.floor(np.amin(x_coords) / grid_size) - padding
        y_min = grid_size * math.floor(np.amin(y_coords) / grid_size) - padding
        x_max = grid_size * math.ceil(np.amax(x_coords) / grid_size) + padding
        y_max = grid_size * math.ceil(np.amax(y_coords) / grid_size) + padding

        width = x_max - x_min
        height = y_max - y_min

        if width < padding * 4 or height < padding * 4:
            raise ValueError(
                "Total padding should not be larger than the original shape of the data"
            )

        bounds = rasterio.coords.BoundingBox(x_min, y_min, x_max, y_max)
        transform = rasterio.transform.from_bounds(
            x_min, y_max, x_max, y_min, width / grid_size, height / grid_size
        )
        window = rasterio.windows.Window(0, 0, width, height)

        return transform, window, bounds

    transform, window, _ = get_transform_from_coords(
        x_coords, y_coords, grid_size=bin_size
    )
    averages, counts = area_bin(values, x_coords, y_coords, transform, window)

    out_avg = np.transpose(np.fliplr(averages))
    out_counts = np.transpose(np.fliplr(counts))

    return transform, out_counts, out_avg


def area_bin(
    values: NDArray,
    x_coords: NDArray,
    y_coords: NDArray,
    transform: rasterio.Affine,
    window: rasterio.windows.Window,
) -> Tuple[NDArray, NDArray]:
    """
    Bins data into a square grid.

    Parameters:
        values: Array of scalar values at each location
        x_coords: Array of x locations
        y_coords: Array of y locations
        bin_size: size of square bins in meters

    Returns:
        (averages, counts, transform): Average of the values contained in each bin,
            and a count of the values contained in each bin.
    """

    # determine grid sizing from a window of dimension 1x1
    w = rasterio.windows.Window(0, 0, 1, 1)
    left, bottom, right, top = rasterio.windows.bounds(w, transform)
    x_bin_size = abs(right - left)
    y_bin_size = abs(top - bottom)

    left, bottom, right, top = rasterio.windows.bounds(window, transform)

    # bin values into a square meter grid and compute averages
    x_bins = np.arange(left, right + x_bin_size, step=x_bin_size)
    y_bins = np.arange(bottom, top + y_bin_size, step=y_bin_size)
    bins = (x_bins, y_bins)

    counts, _, _ = np.histogram2d(x_coords, y_coords, bins=bins)
    value_totals, _, _ = np.histogram2d(x_coords, y_coords, bins=bins, weights=values)

    with np.errstate(divide="ignore", invalid="ignore"):
        averages = np.nan_to_num(value_totals / counts, posinf=0, copy=False)

    return averages, counts



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS ldst table using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    String,
)
from sqlalchemy.orm import mapped_column, relationship

from vipersci.vis.db import Base


class LDST(Base):
    """An object to represent LDST hypothesis info for VIS use."""

    __tablename__ = "ldst"

    id = mapped_column(String, primary_key=True)

    description = mapped_column(
        String, nullable=False, doc="A short descriptive name for this LDST hypothesis."
    )

    image_requests = relationship(
        "ImageRequest",
        secondary="junc_image_request_ldst",
        back_populates="ldst_hypotheses",
        viewonly=True,
    )
    image_request_associations = relationship(
        "JuncImageRequestLDST", back_populates="ldst"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Converts the vertices in a .triN file to a GeoCSV or GeoPackage file.

A .tri10 file is a whitespace separated value file where each line contains ten
elements which describe a triangular facet.  The first nine elements are the
X, Y, and Z coordinates of the verticies, in this order:
x1 y1 z1 x2 y2 z2 x3 y3 z3

The last element is a data value of some kind representing the value of
"the facet".  So for a depth to ice file, that value would be the depth in
meters.

The depth-to-ice .tri10 files sometimes have a -1 to indicate surface ice
(instead of zero).

These files can sometimes have more than 10 elements, and these 'triN' files
have additional columns of data values that should be associated with the
facet in the output file.

This file format is not particularly a standard, but is simple to process.
"""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging

# import csv
from pathlib import Path

import geopandas
from pyproj import CRS, Transformer

# from osgeo import ogr, osr
from shapely.geometry import Polygon

from vipersci import util

logger = logging.getLogger(__name__)

stere_crs = "+proj=stere +lon_0={} +lat_0={} +R=1737400"
sites = dict(  # lon first, then lat
    spole=stere_crs.format(0, -90),
    npole=stere_crs.format(0, 90),
    nobile=stere_crs.format(31.1492746341015, -85.391176037601),
    # the Haworth DEMs are polar stereographic, apparently.
    haworth=stere_crs.format(0, -90),
    viper=stere_crs.format(31.6218, -85.42088),
)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-o", "--output", help="Optional name of output file.", required=False
    )
    parser.add_argument(
        "-c",
        "--csv",
        action="store_true",
        required=False,
        help="Will output a GeoCSV instead of a GeoPackage file.",
    )
    parser.add_argument(
        "-s",
        "--site",
        choices=sites.keys(),
        required=False,
        help="Specifying a site, sets the latitude and longitude for the "
        "center of projection.",
    )
    parser.add_argument(
        "--s_srs",
        default="+proj=cart +a=1737400 +b=1737400",
        help="The source CRS or SRS of the .tri10 data. Default: %(default)s",
    )
    parser.add_argument(
        "--t_srs",
        help="The CRS or SRS of the output file.  The various -s options are "
        "short-hands for this.",
    )
    parser.add_argument(
        "--value_names",
        default="Depth (m)",
        help="This text will be used as the title of the data value field in "
        "the output file.  If there are commas, this text is split on "
        "the commas and whitespace stripped, to produce a list of value "
        "fields in the output. Default: %(default)s",
    )
    parser.add_argument(
        "--value_columns",
        default="9",
        help="This should be a comma-separated list of integers (or just one) "
        "specifying which columns from the .tri file get the names from "
        "-v.  The first 9 columns (zero is the first) have facet "
        "coordinates, so typically this value (or list) starts at 9. "
        "Default: %(default)s",
    )
    parser.add_argument(
        "--value_file",
        help="If provided another whitespace-separated file will be read in, "
        "and the value(s) from the column(s) indicated by --value_columns "
        "will be used as the value(s).",
    )
    parser.add_argument(
        "--remove_facets",
        type=float,
        help="If provided, any facets with this value will not be included in the "
        "output GeoPackage.  Ignored if there is more than one value_column.",
    )
    parser.add_argument(
        "--replace_with_zero",
        help="If there is a value that should be replaced with zero, it can "
        "be provided here.  If the --value_name is 'Depth (m)' this will "
        "automatically be set to '-1'.  If you want to override that, "
        "you can set '--replace_with_zero 0'.",
    )
    parser.add_argument(
        "--keep_z",
        action="store_true",
        help="Will retain the Z value in the output geometries, otherwise "
        "collapses to 2D geometries.",
    )
    parser.add_argument(
        "--keep_all_facets",
        action="store_true",
        help="Will retain all of the triangular facets of the original .tri10 "
        "file, which will result in a very large file.  By default, "
        "adjoining facets with identical values will be merged into "
        "larger polygons.",
    )
    parser.add_argument("file", help=".tri10 file")
    return parser


def arg_checks(args):
    value_keys = list(map(lambda x: x.strip(), args.value_names.strip().split(",")))
    col_idxs = list(
        map(lambda x: int(x.strip()), args.value_columns.strip().split(","))
    )

    if not args.keep_all_facets and len(value_keys) > 1:
        raise ValueError(
            "In order to dissolve facets into larger polygons, there can "
            "be only one --value_column.  Either reduce to one, or set "
            "--keep_all_facets."
        )

    if args.site is not None:
        t_srs = sites[args.site]
    else:
        if args.t_srs is None:
            raise ValueError(
                "Neither a site (-s) nor a target SRS (--t_srs) was provided."
            )
        else:
            t_srs = args.t_srs

    if args.output is None:
        outfile = Path(args.file).with_suffix(".gpkg")
    else:
        outfile = Path(args.output)

    if "Depth (m)" in args.value_names:
        replace_with_zero = -1
    else:
        replace_with_zero = args.replace_with_zero

    return value_keys, col_idxs, t_srs, outfile, replace_with_zero


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    try:
        (value_keys, col_idxs, t_srs, outfile, replace_with_zero) = arg_checks(args)
    except ValueError as err:
        parser.error(str(err))

    s_crs = CRS(args.s_srs)
    t_crs = CRS(t_srs)
    transformer = Transformer.from_crs(s_crs, t_crs)

    values = {k: [] for k in value_keys}

    polys = list()
    with open(args.file, "r") as f:
        logger.info(f"Reading vertices from {args.file}")
        for line in f:
            tokens = line.split()

            if len(col_idxs) == 1 and float(tokens[col_idxs[0]]) == args.remove_facets:
                continue

            poly = vertexes_to_poly(transformer, tokens[:9], args.keep_z)
            polys.append(poly)

            for i, col in enumerate(col_idxs):
                values[value_keys[i]].append(
                    replace_with(0, replace_with_zero, tokens[col])
                )

    if args.value_file is not None:
        values = {k: [] for k in value_keys}  # Re-initialize and empty.
        with open(args.value_file, "r") as vf:
            for line in vf:
                tokens = line.split()
                for i, col in enumerate(col_idxs):
                    values[value_keys[i]].append(
                        replace_with(0, replace_with_zero, tokens[col])
                    )

        if len(values[value_keys[0]]) != len(polys):
            parser.error(
                "The provided value_file has a different number of entries "
                "than the provided .tri file with facet vertices."
            )

    values["geometry"] = polys
    gdf = geopandas.GeoDataFrame(values, crs=t_crs)
    logger.info(gdf.describe())

    if not args.keep_all_facets:
        logger.info("Dissolving polygons.")
        gdf = gdf.dissolve(by=value_keys[0])

    if args.csv:
        gdf.to_file(outfile.with_suffix(".csv"), driver="CSV")
    else:
        logger.info(f"Writing {outfile}")
        gdf.to_file(outfile, driver="GPKG")

    # if args.csv:
    #     with open(args.file, "r") as f:
    #         with open(outfile.with_suffix(".csv"), "w", newline="") as csvfile:
    #             writer = csv.writer(csvfile)
    #             writer.writerow(["WKT", args.value_name])
    #             for line in f:
    #                 tokens = line.split()
    #                 writer.writerow([get_wkt_value(transformer, tokens)])
    # else:
    #     # Write GeoPackage
    #     driver = ogr.GetDriverByName("GPKG")
    #     datasource = driver.CreateDataSource(str(outfile))
    #     srs = osr.SpatialReference()
    #     srs.ImportFromWkt(pstere.to_wkt())
    #     layer = datasource.CreateLayer("depth", srs, ogr.wkbPolygon25D)
    #     layer.CreateField(ogr.FieldDefn(args.value_name, ogr.OFTReal))

    #     with open(args.file, "r") as f:
    #         for line in f:
    #             tokens = line.split()
    #             wkt, value = get_wkt_value(transformer, tokens)
    #             feature = ogr.Feature(layer.GetLayerDefn())
    #             feature.SetField(args.value_name, value)
    #             feature.SetGeometry(ogr.CreateGeometryFromWkt(wkt))
    #             layer.CreateFeature(feature)

    #     # This closes and saves the data source
    #     del datasource

    return


def vertexes_to_poly(transformer, tokens: list, z=True):
    in_m = list(map(lambda x: float(x) * 1000, tokens))

    v1 = transformer.transform(in_m[0], in_m[1], in_m[2])
    v2 = transformer.transform(in_m[3], in_m[4], in_m[5])
    v3 = transformer.transform(in_m[6], in_m[7], in_m[8])

    if z:
        poly = Polygon([v1, v2, v3])
    else:
        poly = Polygon([v1[:-1], v2[:-1], v3[:-1]])

    if poly.area == 0:
        raise ValueError(f"This polygon has zero area: {poly}")

    return poly


def replace_with(replacement_val, replacement_check, value):
    if float(value) == replacement_check:
        return replacement_val
    else:
        return float(value)


# def get_wkt_value(transformer, tokens: list):
#     in_m = list(map(lambda x: float(x) * 1000, tokens[:9]))
#
#     v1 = transformer.transform(in_m[0], in_m[1], in_m[2])
#     v2 = transformer.transform(in_m[3], in_m[4], in_m[5])
#     v3 = transformer.transform(in_m[6], in_m[7], in_m[8])
#
#     wkt = make_wkt(v1, v2, v3)
#     if float(tokens[9]) == -1:
#         value = 0
#     else:
#         value = tokens[9]
#
#     return wkt, value


# def make_wkt(v1: list, v2: list, v3: list) -> str:
#     c_list = ", ".join(
#         [
#             " ".join(map(str, v1)),
#             " ".join(map(str, v2)),
#             " ".join(map(str, v3)),
#             " ".join(map(str, v1))
#         ]
#     )
#     return (f"POLYGON Z (({c_list}))")



"""Creates PDS VIS Pano Products.

This module builds a VIS panorama data products from VIS pano record data.  At this
time, a Pano Product is produced from VIS Image Records. That metadata is the basis for
production of the PDS4 XML label file.

In order to perform the lookup of Image Records and Pano Records requires use
of the database.

For now, this program still has a variety of hard-coded elements,
that will eventually be extracted from telemetry.

The command-line version is primarily to aide testing.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import timezone
import json
import logging
from typing import Union
from pathlib import Path

import numpy as np
import numpy.typing as npt
from sqlalchemy import create_engine, select
from sqlalchemy.orm import Session

import vipersci
from vipersci.vis.db.pano_records import PanoRecord
from vipersci.vis.create_image import tif_info
from vipersci.pds import pid as pds
from vipersci import util
from vipersci.vis.pds import lids, write_xml

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-d",
        "--dburl",
        required=True,
        help="Database with a pano_products table which will be written to. "
        "If not given, no database will be written to.  Example: "
        "postgresql://postgres:NotTheDefault@localhost/visdb",
    )
    parser.add_argument(
        "-t",
        "--template",
        default="pano-template.xml",
        help="Genshi XML file template.  Will default to the pano-template.xml "
        "file distributed with the module.",
    )
    parser.add_argument(
        "--tiff",
        type=Path,
        help="Optional pre-existing TIFF file (presumably created by create_image). "
        "This file will be inspected and its information added to the output. ",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument(
        "input",
        help="Product ID (or TIFF file from which a Product ID can be extracted) or a "
        "JSON file containing metadata.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    engine = create_engine(args.dburl)
    with Session(engine) as session:
        try:
            pid = pds.PanoID(args.input)
        except ValueError:
            pid = None

        pr = None
        if pid is None or args.input.endswith(".json"):
            if Path(args.input).exists():
                with open(args.input) as f:
                    pr = PanoRecord(**json.load(f))
            else:
                parser.error(f"The file {args.input} does not exist.")
        else:
            # We got a valid pid, go look it up in the db.
            stmt = select(PanoRecord).where(PanoRecord.product_id == str(pid))
            result = session.scalars(stmt)
            rows = result.all()
            if len(rows) > 1:
                raise ValueError(f"There was more than 1 row returned from {stmt}")
            elif len(rows) == 0:
                raise ValueError(
                    f"No records were returned from the database for Product Id {pid}."
                )

            pr = rows[0]

        if pr is None:
            raise ValueError(f"Could not extract a PanoRecord from {args.input}")

        # If testing via sqlite, ensure datetimes are UTC aware:
        if session.get_bind().name == "sqlite":
            pr.file_creation_datetime = pr.file_creation_datetime.replace(
                tzinfo=timezone.utc
            )
            pr.start_time = pr.start_time.replace(tzinfo=timezone.utc)
            pr.stop_time = pr.stop_time.replace(tzinfo=timezone.utc)

        # I'm not sure where these are coming from, let's hard-code them for now:
        metadata = {
            "mission_phase": "TEST",
        }

        # This allows values in these dicts to override the hard-coded values above.
        metadata.update(label_dict(pr))
        metadata.update(pr.asdict())
        metadata.update(
            {
                "software_name": "vipersci",
                "software_version": vipersci.__version__,
                "software_type": "Python",
                "software_program_name": __name__,
            }
        )
        if metadata["purpose"] is None:
            metadata["purpose"] = "Science"
        else:
            metadata["purpose"] = metadata["purpose"].value.replace("_", " ").title()

    if args.input.endswith(".tif"):
        args.tiff = Path(args.input)

    if args.tiff is None:
        # Make up some values:
        metadata["file_byte_offset"] = 0
        metadata["file_data_type"] = "UnsignedLSB2"
    else:
        t_info = tif_info(args.tiff)

        for k, v in t_info.items():
            if hasattr(metadata, k) and metadata[k] != v:
                raise ValueError(
                    f"The value of {k} in the metadata ({metadata[k]}) does not match "
                    f"the value ({v}) in the image ({args.tiff})"
                )
        metadata.update(t_info)

    write_xml(metadata, args.template, args.output_dir)

    return


def label_dict(pr: PanoRecord):
    """Returns a dictionary suitable for label generation."""
    d = dict(
        lid=f"{lids['bundle']}:data_derived:{pr.product_id}",
        mission_lid=lids["mission"],
        sc_lid=lids["spacecraft"],
        instruments=[],
        source_product_lidvids=[],
        source_product_type="data_to_raw_source_product",
    )

    instruments_dict = {}
    for ir in pr.image_records:
        _inst = ir.instrument_name.lower().replace(" ", "_")
        instruments_dict[ir.instrument_name] = {
            "name": ir.instrument_name,
            "lid": f"{lids['spacecraft']}.{_inst}",
        }
        # todo: need to figure out how to set version id here
        d["source_product_lidvids"].append(  # type: ignore
            f"urn:nasa:pds:viper_vis:data_raw:{ir.product_id}::99.99"
        )

    for inst in instruments_dict.values():
        d["instruments"].append(inst)  # type: ignore

    return d



#!/usr/bin/env python
"""This module has tests for the vis.pds.labelmaker functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timezone
from pathlib import Path
from textwrap import dedent
import unittest
from unittest.mock import call, create_autospec, mock_open, patch
import xml.etree.ElementTree as ET

from genshi.template import Template
import pandas as pd

import vipersci.pds.labelmaker as lm


class TestFunctions(unittest.TestCase):
    def setUp(self):
        self.xmltext = dedent(
            """\
        <?xml version="1.0" encoding="UTF-8"?>
        <?xml-model href="https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.sch"
          schematypens="http://purl.oclc.org/dsdl/schematron"?>
        <Product_Collection
          xmlns="http://pds.nasa.gov/pds4/pds/v1"
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
          xsi:schemaLocation="http://pds.nasa.gov/pds4/pds/v1
          https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.xsd">
        <Identification_Area>
            <logical_identifier>urn:nasa:pds:dummy_lid</logical_identifier>
            <version_id>1.0</version_id>
          <Modification_History>
            <Modification_Detail>
              <modification_date>2023-10-28</modification_date>
              <version_id>0.1</version_id>
              <description>Bogus testing version</description>
            </Modification_Detail>
          </Modification_History>
        </Identification_Area>
        <Context_Area>
          <Time_Coordinates>
            <start_date_time>2023-10-26T20:00:00Z</start_date_time>
            <stop_date_time>2023-10-26T20:00:00.000511Z</stop_date_time>
          </Time_Coordinates>
          <Primary_Result_Summary>
            <purpose>Engineering</purpose>
            <processing_level>Raw</processing_level>
          </Primary_Result_Summary>
          <Investigation_Area>
            <name>MISSIONNAME</name>
            <type>Mission</type>
            <Internal_Reference>
              <lid_reference>urn:nasa:pds:missionname</lid_reference>
              <reference_type>collection_to_investigation</reference_type>
            </Internal_Reference>
          </Investigation_Area>
          <Observing_System>
            <Observing_System_Component>
              <name>MISSIONNAME</name>
              <type>Host</type>
              <Internal_Reference>
                <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.missionname</lid_reference>
                <reference_type>is_instrument_host</reference_type>
              </Internal_Reference>
            </Observing_System_Component>
            <Observing_System_Component>
                <name>NavCam Left</name>
                <type>Instrument</type>
                <Internal_Reference>
                    <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.missionname.navcam_left</lid_reference>
                    <reference_type>is_instrument</reference_type>
                </Internal_Reference>
            </Observing_System_Component><Observing_System_Component>
                <name>NavCam Right</name>
                <type>Instrument</type>
                <Internal_Reference>
                    <lid_reference>urn:nasa:pds:context:instrument_host:spacecraft.missionname.navcam_right</lid_reference>
                    <reference_type>is_instrument</reference_type>
                </Internal_Reference>
            </Observing_System_Component>
          </Observing_System>
          <Target_Identification>
            <name>Moon</name>
            <type>Satellite</type>
            <Internal_Reference>
              <lid_reference>urn:nasa:pds:context:target:satellite.earth.moon</lid_reference>
              <reference_type>collection_to_target</reference_type>
            </Internal_Reference>
          </Target_Identification>
        </Context_Area>
        <Collection>
          <collection_type>Data</collection_type>
        </Collection>
        <File_Area_Inventory>
          <File>
            <file_name>collection_data_raw.csv</file_name>
            <creation_date_time>2023-11-02T23:12:59.083415Z</creation_date_time>
          </File>
          <Inventory>
            <offset unit="byte">0</offset>
            <parsing_standard_id>PDS DSV 1</parsing_standard_id>
            <records>4</records>
            <record_delimiter>Carriage-Return Line-Feed</record_delimiter>
            <field_delimiter>Comma</field_delimiter>
          </Inventory>
        </File_Area_Inventory>
        </Product_Collection>
        """
        )

    def test_get_lidvidfile(self):
        mock_path = create_autospec(Path)
        mock_path.read_text.return_value = dedent(
            """\
            <?xml version="1.0" encoding="UTF-8"?>
            <?xml-model
              href="https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.sch"
              schematypens="http://purl.oclc.org/dsdl/schematron"?>
            <Product_Observational
              xmlns="http://pds.nasa.gov/pds4/pds/v1"
              xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
              xsi:schemaLocation="http://pds.nasa.gov/pds4/pds/v1
              https://pds.nasa.gov/pds4/pds/v1/PDS4_PDS_1K00.xsd">
            <Identification_Area>
                <logical_identifier>urn:nasa:pds:dummy_lid</logical_identifier>
                <version_id>1.0</version_id>
            </Identification_Area>
            <File_Area_Observational>
              <File>
                <file_name>product.name</file_name>
              </File>
            </File_Area_Observational>
            </Product_Observational>
            """
        )

        truth = {
            "lid": "urn:nasa:pds:dummy_lid",
            "vid": "1.0",
            "productfile": "product.name",
        }

        d = lm.get_lidvidfile(mock_path)

        self.assertEqual(truth, d)

    def test_get_common_label_info(self):
        root = ET.fromstring(self.xmltext)
        host_lid = "urn:nasa:pds:context:instrument_host:spacecraft.missionname"
        truth = {
            "lid": "urn:nasa:pds:dummy_lid",
            "vid": "1.0",
            "start_date_time": datetime(2023, 10, 26, 20, 00, 00, tzinfo=timezone.utc),
            "stop_date_time": datetime(
                2023, 10, 26, 20, 00, 00, 511, tzinfo=timezone.utc
            ),
            "investigation_name": "MISSIONNAME",
            "investigation_type": "Mission",
            "investigation_lid": "urn:nasa:pds:missionname",
            "host_name": "MISSIONNAME",
            "host_lid": host_lid,
            "target_name": "Moon",
            "target_type": "Satellite",
            "target_lid": "urn:nasa:pds:context:target:satellite.earth.moon",
            "instruments": {
                "NavCam Left": f"{host_lid}.navcam_left",
                "NavCam Right": f"{host_lid}.navcam_right",
            },
            "purposes": [
                "Engineering",
            ],
            "processing_levels": [
                "Raw",
            ],
        }
        d = lm.get_common_label_info(root, area="pds:Context_Area")
        self.assertEqual(truth, d)

    def test_gather_info(self):
        df = pd.DataFrame(
            data={
                "vid": ["0.1", "0.1", "1.0"],
                "instruments": [
                    {
                        "urn:nasa:pds:ncl": "NavCam Left",
                        "urn:nasa:pds:ncr": "NavCam Right",
                    },
                    {
                        "urn:nasa:pds:ncl": "NavCam Left",
                        "urn:nasa:pds:ncr": "NavCam Right",
                    },
                    {"urn:nasa:pds:acl": "AftCam Left"},
                ],
                "purposes": [
                    [
                        "Engineering",
                    ],
                    [
                        "Science",
                    ],
                    [
                        "Science",
                    ],
                ],
                "processing_levels": [
                    [
                        "Raw",
                    ],
                    [
                        "Raw",
                    ],
                    [
                        "Derived",
                    ],
                ],
                "start_date_time": [
                    datetime(2023, 10, 1, 20, 00, 00, tzinfo=timezone.utc),
                    datetime(2023, 10, 15, 20, 00, 00, tzinfo=timezone.utc),
                    datetime(2023, 11, 1, 20, 00, 00, tzinfo=timezone.utc),
                ],
                "stop_date_time": [
                    datetime(2023, 10, 1, 20, 00, 1, tzinfo=timezone.utc),
                    datetime(2023, 10, 15, 20, 00, 1, tzinfo=timezone.utc),
                    datetime(2023, 11, 1, 20, 00, 1, tzinfo=timezone.utc),
                ],
            }
        )

        truth = {
            "vid": "2.0",
            "instruments": {
                "urn:nasa:pds:ncl": "NavCam Left",
                "urn:nasa:pds:ncr": "NavCam Right",
                "urn:nasa:pds:acl": "AftCam Left",
            },
            "purposes": {"Engineering", "Science"},
            "processing_levels": {"Raw", "Derived"},
            "start_date_time": "2023-10-01T20:00:00Z",
            "stop_date_time": "2023-11-01T20:00:01Z",
        }
        d = lm.gather_info(df, [{"version": "0.1"}, {"version": "2.0"}])
        self.assertEqual(truth, d)

    def test_assert_unique(self):
        self.assertIsNone(lm.assert_unique("a", pd.Series(["a", "a"])))
        self.assertRaises(ValueError, lm.assert_unique, "a", pd.Series(["b", "b"]))
        self.assertRaises(ValueError, lm.assert_unique, "a", pd.Series(["a", "b"]))

    def test_vid_max(self):
        mod_details = [
            {"version": "1.0"},
            {"version": "1.1"},
            {"version": "2.0"},
        ]
        self.assertEqual(lm.vid_max(mod_details), 2.0)
        self.assertEqual(lm.vid_max(mod_details, 2.0), 2.0)
        self.assertRaises(ValueError, lm.vid_max, mod_details, 3.0)

    def test_write_inventory(self):
        m = mock_open()
        labels = [
            {"lid": "lid1", "vid": "1.0"},
            {"lid": "lid2", "vid": "1.0"},
        ]
        with patch("vipersci.pds.labelmaker.open", m):
            lm.write_inventory(Path("dummy_path"), labels)

        handle = m()
        handle.write.assert_has_calls(
            [call("P,lid1::1.0\r\n"), call("P,lid2::1.0\r\n")]
        )

    def test_write_xml(self):
        m_path = create_autospec(Path)
        m_tmpl_path = create_autospec(Path)
        m_tmpl = create_autospec(Template)
        with patch(
            "vipersci.pds.labelmaker.MarkupTemplate", return_value=m_tmpl
        ) as m_markup:
            lm.write_xml({"lid": "dummy_lid", "vid": "1.0"}, m_path, m_tmpl_path)

        m_tmpl_path.read_text.assert_called_once()
        m_markup.assert_called_once()
        m_tmpl.generate.assert_called_once()
        m_path.write_text.assert_called_once()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Takes input JSON file, and creates an output CSV file.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
import logging
from pathlib import Path

import geopandas as gpd
from shapely.geometry import Point

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-i",
        "--interval",
        type=float,
        default=1.0,
        help="The time interval in seconds to interpolate the provided " "traverse at.",
    )
    parser.add_argument("-o", "--output", type=Path, help="Output CSV file.")
    parser.add_argument(
        "json",
        type=Path,
        help="The JSON file that contains a traverse.",
        # Assumed to be GeoJSON from the Map Server.
    )

    return parser


def main():
    args = arg_parser().parse_args()

    points = list()
    df = gpd.read_file(args.json)
    # ['start_time', 'uuid', 'item_type', 'end_time', 'duration', 'name',
    # 'color', 'text_color', 'activity_type', 'style', 'geometry']

    last_point = Point(0, 0)
    last_time = None
    for row in df.itertuples():
        if row.name != "Driving":
            continue

        # If two consecutive "Driving" segments have the same end and
        # beginning coordinates, but the time is different, that means
        # that time passed while we sat there.
        # print(f"Points: {row.geometry.coords[0]} == {last_point.coords[0]} ")
        # print(f"Times: {row.start_time} != {last_time}")
        if (
            row.geometry.coords[0] == last_point.coords[0]
            and row.start_time != last_time
        ):
            time_intervals = interval_count(last_time, row.start_time, args.interval)
            for s in range(time_intervals):
                t = last_time + (s * args.interval)
                points.append({"time": t, "x": last_point.x, "y": last_point.y})

        time_intervals = interval_count(row.start_time, row.end_time, args.interval)
        for s in range(time_intervals):
            t = row.start_time + (s * args.interval)
            p = row.geometry.interpolate(s / time_intervals, normalized=True)
            points.append({"time": t, "x": p.x, "y": p.y})
        last_time = row.end_time
        last_point = Point(row.geometry.coords[-1])

    with open(args.output, "w", newline="") as csvfile:
        fieldnames = ["time", "x", "y"]
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        for d in points:
            writer.writerow(d)

    return


def interval_count(start, stop, step):
    # This seems too simple, and I feel like there should be a one-liner that
    # does this.  Until there is, this pattern gets used twice, so a function
    # it is.
    total_time = stop - start
    return int(total_time / step)



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS pano_records table using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime

from sqlalchemy import (
    DateTime,
    Enum,
    Float,
    Identity,
    Integer,
    String,
)
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import mapped_column, relationship, validates

from vipersci.pds import Purpose
from vipersci.pds.pid import VISID, PanoID
from vipersci.pds.datetime import isozformat
from vipersci.vis.db import Base
import vipersci.vis.db.validators as vld


class PanoRecord(Base):
    """An object to represent rows in the pano_records table for VIS."""

    # This class is derived from SQLAlchemy's orm.DeclarativeBase
    # which means that it has a variety of class properties that are
    # then swept up into properties on the instantiated object via
    # super().__init__().

    # The table represents many of these objects, so the __tablename__ is
    # plural while the class name is singular.
    __tablename__ = "pano_records"

    # The mapped_column() names below should use "snake_case" for the names that are
    # committed to the database as column names.  Furthermore, those names
    # should be similar, if not identical, to the PDS4 Class and Attribute
    # names that they represent.  Other names are implemented as synonyms.
    # Aside from the leading "id" column, the remainder are in alphabetical order.

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    file_creation_datetime = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        doc="The time at which file_name was created.",
    )
    file_md5_checksum = mapped_column(
        String,
        nullable=False,
        doc="The md5 checksum of the file described by file_path.",
    )
    file_path = mapped_column(
        String,
        nullable=False,
        doc="The absolute path (POSIX style) that contains the Array_2D_Image "
        "that this metadata refers to.",
    )
    # The image_records and image_record_associations allow a many PanoRecord to many
    # ImageRecord relationship.
    image_records = relationship(
        "ImageRecord",
        secondary="junc_image_pano",
        back_populates="pano_records",
        viewonly=True,
    )
    image_record_associations = relationship(
        "JuncImagePano", back_populates="pano_record"
    )
    lines = mapped_column(
        Integer,
        nullable=False,
        doc="The number of lines or rows in the Pano Product image.",
    )
    rover_pan_min = mapped_column(
        Float,
        nullable=False,
        doc="The minimum or leftmost pan angle of this panorama in degrees, where "
        "zero is the rover forward (+x) direction. "
        "This angle is relative to the rover frame.",
    )
    rover_pan_max = mapped_column(
        Float,
        nullable=False,
        doc="The maximum or rightmost pan angle of this panorama in degrees, where "
        "zero is the rover forward (+x) direction.  "
        "This angle is relative to the rover frame.",
    )
    rover_tilt_min = mapped_column(
        Float,
        nullable=False,
        doc="The minimum or lower tilt angle of this panorama in degrees, "
        "where zero is the rover level plane.  "
        "This tilt angle is relative to the rover frame.",
    )
    rover_tilt_max = mapped_column(
        Float,
        nullable=False,
        doc="The maximum or upper tilt angle of this panorama in degrees, "
        "where zero is the rover level plane.  "
        "This tilt angle is relative to the rover frame.",
    )
    _pid = mapped_column(
        "product_id", String, nullable=False, unique=True, doc="The PDS Product ID."
    )
    purpose = mapped_column(
        Enum(Purpose),
        nullable=True,
        doc="This is the value for the PDS "
        "Observation_Area/Primary_Result_Summary/purpose parameter, it "
        "has a restricted set of allowable values.",
    )
    samples = mapped_column(
        Integer,
        nullable=False,
        doc="The number of samples or columns in the Pano Product Image.",
    )
    software_name = mapped_column(String, nullable=False)
    software_version = mapped_column(String, nullable=False)
    software_program_name = mapped_column(String, nullable=False)
    start_time = mapped_column(DateTime(timezone=True), nullable=False)
    stop_time = mapped_column(DateTime(timezone=True), nullable=False)

    def __init__(self, **kwargs):
        # If present, product_id needs some special handling:
        if "product_id" in kwargs:
            pid = PanoID(kwargs["product_id"])
            del kwargs["product_id"]
        else:
            pid = False

        ppargs = dict()
        otherargs = dict()
        for k, v in kwargs.items():
            if k in self.__table__.columns or k in self.__mapper__.synonyms:
                ppargs[k] = v
            else:
                otherargs[k] = v

        # Instantiate early, so that the parent orm_declarative Base can
        # resolve all of the synonyms.
        super().__init__(**ppargs)

        # Ensure product_id consistency
        if pid:
            if "start_time" in kwargs and pid.datetime() != self.start_time:
                raise ValueError(
                    f"The product_id datetime ({pid.datetime()}) and the "
                    f"provided start_time ({kwargs['start_time']}) disagree."
                )

        elif "source_pids" in kwargs:
            source_pids = list(map(VISID, kwargs["source_pids"]))
            instruments = set([p.instrument for p in source_pids])
            inst = "pan"
            if len(instruments) == 1:
                inst = instruments.pop()

            source_pids.sort()
            st = source_pids[0].datetime()
            if self.start_time is None:
                self.start_time = st
            else:
                st = self.start_time

            pid = PanoID(st.date(), st.time(), inst)
        else:
            got = dict()
            for k in (
                "product_id",
                "start_time",
            ):
                v = getattr(self, k)
                if v is not None:
                    got[k] = v

            if "source_pids" in kwargs:
                got["source_pids"] = kwargs["source_pids"]

            raise ValueError(
                "Either product_id must be given, or a list of source product IDs. "
                f"Got: {got}"
            )

        self._pid = str(pid)

        # Is this really a good idea?  Not sure.  This instance variable plus
        # label_dict() and update() allow other key/value pairs to be carried around
        # in this object, which is handy.  If these are well enough known, perhaps
        # they should just be pre-defined properties and not left to chance?
        self.labelmeta = otherargs

        return

    @hybrid_property
    def product_id(self):
        return self._pid

    @product_id.inplace.setter
    def _product_id_setter(self, pid):
        # In this class, the source of product_id information really is what
        # based on the source products, and so this should not be monkeyed with.
        # So at this time, this can only be set when this object is instantiated.
        raise NotImplementedError(
            "product_id cannot be set directly after instantiation."
        )

    @validates(
        "file_creation_datetime",
        "start_time",
        "stop_time",
    )
    def validate_datetime_asutc(self, key, value):
        return vld.validate_datetime_asutc(key, value)

    @validates("purpose")
    def validate_purpose(self, key, value: str):
        return vld.validate_purpose(value)

    def asdict(self):
        d = {}

        for c in self.__table__.columns:
            if isinstance(getattr(self, c.name), datetime):
                d[c.name] = isozformat(getattr(self, c.name))
            else:
                d[c.name] = getattr(self, c.name)

        if hasattr(self, "labelmeta"):
            d.update(self.labelmeta)

        return d

    @classmethod
    def from_xml(cls, text: str):
        """
        Returns an instantiated PanoProduct object from parsing the provided *text*
        as XML.
        """
        raise NotImplementedError()

    def update(self, other):
        for k, v in other.items():
            if k in self.__table__.columns or k in self.__mapper__.synonyms:
                setattr(self, k, v)
            else:
                self.labelmeta[k] = v



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Takes GPKG file nominally created by tri2gpkg.py, and creates a simplified
geometry of the different ISR types.
"""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging

import geopandas as gp
import pandas as pd
import shapely

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-o", "--out", help="Output GeoPackage filename")
    parser.add_argument(
        "gpkg", help="The geospatial file containing polygon geometries."
    )

    return parser


def main():
    args = arg_parser().parse_args()

    df = gp.read_file(args.gpkg)

    print("Read in file.")

    if (df.has_z).any():
        df["geometry"] = df.apply(
            lambda row: shapely.ops.transform(
                lambda x, y, z=None: (x, y), row["geometry"]
            ),
            axis=1,
            result_type="expand",
        )

        print("Converted geometries to 2D.")
    # print(df)

    # based on VIPER-SCI-PLN-002:
    # 0 m: Surficial
    # 0 - 0.5 m: Shallow
    # 0.5 - 1 m: Deep
    # > 1 m: Dry
    df["category"] = pd.cut(
        df["Depth (m)"].astype(float),
        (-1, 0, 0.5, 1, 10),
        labels=["Surficial", "Shallow", "Deep", "Dry"],
    ).astype(str)

    print("Added categories.")

    dissolved = df.dissolve(by="category")

    print("Dissolved geometries.")

    dissolved.to_file(args.out, driver="GPKG")



"""Creates a PDS4 Collection XML file from the provided XML labels.
"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from datetime import datetime, timezone
import logging
from pathlib import Path
import xml.etree.ElementTree as ET

import pandas as pd
import yaml

from vipersci.pds.datetime import isozformat
from vipersci.pds.labelmaker import (
    assert_unique,
    gather_info,
    get_common_label_info,
    vid_max,
    write_inventory,
    write_xml,
)
from vipersci.pds.xml import ns
from vipersci import util

logger = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser("collection", help=__doc__)

    parser.add_argument(
        "-c", "--config", type=Path, help="YAML file with configuration parameters."
    )
    parser.add_argument(
        "--csv",
        type=Path,
        help="The collections CSV file.  If not given, one will be generated.",
    )
    parser.add_argument(
        "-t", "--template", type=Path, required=True, help="Genshi XML file template. "
    )
    parser.add_argument(
        "labelfiles",
        type=Path,
        nargs="+",
        help="Path(s) to all XML label files to be included in the output collection "
        "file.",
    )
    parser.set_defaults(func=main)
    return parser


def main(args):
    util.set_logger(args.verbose)

    d = yaml.safe_load(args.config.read_text())

    name = d["collection_lid"].split(":")[-1]
    outpath = Path(f"collection_{name}.xml")
    if outpath.exists():
        raise FileExistsError(f"The file {outpath} already exists.")

    labelinfo = []
    for labelpath in args.labelfiles:
        logger.info(f"Reading {labelpath}")
        labelinfo.append(get_label_info(labelpath))

    d.update(check_and_derive(d, labelinfo))

    if args.csv is None:
        csv_path = Path(f"collection_{name}.csv")
        if csv_path.exists():
            raise FileExistsError(f"The file {csv_path} already exists.")

        write_inventory(csv_path, labelinfo)
        logger.info(f"Wrote {csv_path}")
        d["number_of_records"] = len(labelinfo)
    else:
        csv_path = args.csv
        d["number_of_records"] = len(str.splitlines(args.csv.read_text()))

    d["collection_csv"] = csv_path
    d["file_creation_datetime"] = isozformat(
        datetime.fromtimestamp(csv_path.stat().st_mtime, timezone.utc)
    )

    write_xml(d, outpath, args.template)
    logger.info(f"Wrote {outpath}")

    return


def check_and_derive(config: dict, labelinfo: list):
    df = pd.DataFrame(labelinfo)

    check = {
        "Browse": ("collection_lid",),
        "Data": (
            "collection_lid",
            "investigation_name",
            "investigation_type",
            "investigation_lid",
            "host_name",
            "host_lid",
            "target_name",
            "target_type",
            "target_lid",
        ),
        "Document": ("collection_lid",),
    }

    # Check consistency for gathered labels:
    for x in check[config["collection_type"]]:
        assert_unique(config[x], df[x])

    # Generate values from gathered labels:
    if config["collection_type"] == "Data":
        d = gather_info(df, config["modification_details"])
    elif config["collection_type"] in ("Document", "Browse"):
        d = {
            "vid": str(
                vid_max(config["modification_details"], pd.to_numeric(df["vid"]).max())
            )
        }
    else:
        raise NotImplementedError(
            f"Do not have a strategy for {config['collection_type']} collection types."
        )

    return d


def get_label_info(path: Path) -> dict:
    root = ET.fromstring(path.read_text())

    if root.find("pds:Observation_Area", ns) is not None:
        d = get_common_label_info(root, "pds:Observation_Area")
    else:
        d = get_common_label_info(root, None)

    d["collection_lid"] = ":".join(d["lid"].split(":")[:-1])
    return d



#!/usr/bin/env python
"""This module has tests for the vipersci.vis.create_tif functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
from pathlib import Path
import unittest
from unittest.mock import patch

import vipersci.pds.pid as pds
import vipersci.vis.create_tif as ct


class TestParser(unittest.TestCase):
    def test_arg_parser(self):
        p = ct.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        self.assertRaises(SystemExit, p.parse_args)
        d = vars(p.parse_args(["-p", "product_id_goes_here", "dummy.png"]))
        self.assertIn("product_id", d)
        self.assertIn("output_dir", d)
        self.assertIn("image", d)


class TestMain(unittest.TestCase):
    @patch("vipersci.vis.create_tif.write_tiff")
    @patch("vipersci.vis.create_tif.imread")
    def test_main(self, m_imread, m_writetiff):
        fn = "dummy.png"
        pid = pds.VISID("231125-140902-ncl-c")
        pa_ret_val = ct.arg_parser().parse_args(
            [
                "--product_id",
                str(pid),
                fn,
            ]
        )
        with patch("vipersci.vis.create_tif.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val
            ct.main()
            m_imread.assert_called_once_with(fn)
            m_writetiff.assert_called_once()
            self.assertEqual(m_writetiff.call_args[0][0], pid)
            self.assertEqual(m_writetiff.call_args[0][2], Path.cwd())



#!/usr/bin/env python
"""This module has tests for the image_stats module."""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest

from vipersci.vis.db import image_stats as rs


class TestRawStats(unittest.TestCase):
    def setUp(self):
        self.d = dict(
            blur=0.5,
            mean=2096.5,
            std=1027.3,
        )

    def test_init(self):
        stats = rs.ImageStats(**self.d)
        self.assertEqual(2096.5, stats.mean)



#!/usr/bin/env python
# coding: utf-8

"""Defines the VIS image_tags table using the SQLAlchemy ORM."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from sqlalchemy import (
    Identity,
    Integer,
    String,
)
from sqlalchemy.orm import mapped_column, relationship

from vipersci.vis.db import Base


class ImageTag(Base):
    """An object to represent tags that can be applied to VIS Image Records."""

    __tablename__ = "image_tags"

    id = mapped_column(Integer, Identity(start=1), primary_key=True)
    name = mapped_column(String, nullable=False, doc="The name of the tag.")

    image_records = relationship(
        "ImageRecord",
        secondary="junc_image_record_tags",
        back_populates="image_tags",
        viewonly=True,
    )
    image_record_associations = relationship(
        "JuncImageRecordTag", back_populates="image_tag"
    )


# Unlike other tables, the image_tags table should not have arbitrary contents, and the
# set of tags should be uniform and constant.  The taglist below is the canonical list
# of tag names to be loaded into this table.
# Only add new tags to the bottom of the list.
taglist = [
    "Over Exposed",
    "Under Exposed",
    "Compression Artifacts",
    "Missing row(s) of pixels",
    # Image header damaged in downlink?  Not sure how we'd know.
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Takes a geospatial vector file and creates either positive or negative
buffered versions of the regions.
"""

# Copyright 2021-2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse

import geopandas as gp
from shapely.geometry import Polygon, MultiPolygon
from shapely.geometry.collection import GeometryCollection
from shapely.ops import unary_union


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-b",
        "--buffer",
        type=float,
        help="A positive distance has an effect of dilation; a negative "
        "distance, erosion.",
    )
    parser.add_argument("-o", "--out", help="Output GeoPackage filename.")
    parser.add_argument(
        "gpkg",
        help="The geospatial file with regions to evaluate.  This file "
        "Must have only four geometries, with a category fields of "
        "Deep, Dry, Shallow, and Surficial.",
    )

    return parser


def main():
    isr_order = ("Surficial", "Shallow", "Deep", "Dry")
    index_col = "category"

    parser = arg_parser()
    args = parser.parse_args()

    df = gp.read_file(args.gpkg)
    print("Read in file.")

    categories = df.set_index(index_col)

    if args.buffer == 0:
        parser.error(
            "The buffer value is zero, which would result in no change to the map."
        )
    elif args.buffer < 0:
        # Since we are shrinking the geometries, we must first put them
        # together.
        expanded = list()
        for cat in categories.index:
            if cat == isr_order[0]:
                # The first geometry doesn't combine
                expanded.append(categories.loc[isr_order[0]]["geometry"])
            else:
                # The rest do:
                i = isr_order.index(cat)
                expanded.append(
                    categories.loc[list(isr_order[0 : i + 1])]
                    .dissolve()["geometry"]
                    .iloc[0]
                )

        categories.set_geometry(expanded, inplace=True)

    buffered = categories.buffer(args.buffer)

    print("Buffering complete.")

    for cutting in isr_order[:3]:  # The last geom won't reduce any others.
        for geom in isr_order[isr_order.index(cutting) + 1 :]:
            buffered.loc[geom] = buffered.loc[geom].difference(buffered.loc[cutting])

    # The differencing may have resulted in a GeometryCollection instead
    # of a clean MultiPolygon.
    cleaned = buffered.apply(clean)

    print("Differencing complete.")

    cleaned.to_file(args.out, driver="GPKG")


def clean(geometry):
    if isinstance(geometry, GeometryCollection):
        polys = list()
        for geom in geometry:
            if isinstance(geom, (Polygon, MultiPolygon)):
                polys.append(geom)
        return unary_union(polys)

    else:
        return geometry



"""Creates Browse VIS PDS Products.

This module builds "browse" VIS data products from VIS Image Products.
"""

# Copyright 2022-2024, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
from datetime import datetime, timezone
import logging
from typing import Union
from pathlib import Path
import xml.etree.ElementTree as ET

import numpy as np
import numpy.typing as npt
from skimage.io import imread, imsave
from skimage.exposure import equalize_adapthist, rescale_intensity
from skimage.transform import resize

from vipersci.vis.pds import write_xml
from vipersci.pds.labelmaker import get_lidvidfile
from vipersci.pds.datetime import isozformat
from vipersci.pds.xml import find_text, ns
from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-t",
        "--template",
        default="browse-template.xml",
        help="Genshi XML file template.  Will default to the browse-template.xml "
        "file distributed with the module.",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        type=Path,
        default=Path.cwd(),
        help="Output directory for label. Defaults to current working directory.  "
        "Output file names are fixed based on product_id, and will be over-written.",
    )
    parser.add_argument(
        "input",
        type=Path,
        help="PDS XML label file of product to create a browse product from.",
    )
    return parser


def main():
    parser = arg_parser()
    args = parser.parse_args()
    util.set_logger(args.verbose)

    metadata = get_product_info(args.input)

    if metadata["productfile"] is None:
        raise ValueError(f"Could not fine a file_name in {args.input}")

    image_path = args.input.parent / metadata["productfile"]

    try:
        image = rescale_intensity(
            equalize_adapthist(imread(image_path)), in_range="image", out_range="uint8"
        )
    except FileNotFoundError as err:
        parser.error(str(err))

    metadata["source_lidvid"] = metadata["lid"] + "::" + metadata["vid"]
    metadata["source_product_type"] = (
        f"data_to_{metadata['type'].lower()}_source_product"
    )

    # Adjust LID
    lid_tokens = metadata["lid"].split(":")
    lid_tokens[4] = "browse"
    metadata["product_id"] = lid_tokens[5] + "-browse"
    lid_tokens[5] = metadata["product_id"]
    metadata["lid"] = ":".join(lid_tokens)

    # Adjust title
    title_tokens = metadata["title"].split(" - ")
    title_words = title_tokens[0].split()
    title_words.insert(-1, "browse")
    metadata["title"] = " ".join(title_words) + f" - {metadata['product_id']}"

    # Scale down image to be no larger than 1024 pixels in the maximum direction
    max_dim = max(np.shape(image))
    if max_dim > 1024:
        scale = max_dim / 1024
        new_shape = tuple(int(x / scale) for x in np.shape(image))
        image = rescale_intensity(
            resize(image, new_shape), in_range="image", out_range="uint8"
        )

    metadata["file_path"] = metadata["product_id"] + ".png"
    out_im_path = args.output_dir / metadata["file_path"]
    imsave(out_im_path, image, check_contrast=False)

    metadata["file_creation_datetime"] = isozformat(
        datetime.fromtimestamp(out_im_path.stat().st_mtime, timezone.utc)
    )
    metadata["file_md5_checksum"] = util.md5(out_im_path)

    write_xml(metadata, args.template, args.output_dir)

    return


def get_product_info(xml_path: Path):
    d = get_lidvidfile(xml_path)

    label = ET.fromstring(xml_path.read_text())
    d["title"] = find_text(label, "./pds:Identification_Area/pds:title")
    d["type"] = find_text(
        label, "./pds:Observation_Area/pds:Primary_Result_Summary/pds:processing_level"
    )
    d["modification_details"] = get_modification_details(label)
    return d


def get_modification_details(element_tree: ET.Element):
    mod_history = list()
    for mod_detail in element_tree.findall(".//pds:Modification_Detail", ns):
        d = dict()
        for key in ("modification_date", "version_id", "description"):
            d[key] = find_text(mod_detail, f"pds:{key}")
        mod_history.append(d)

    return mod_history



#!/usr/bin/env python
# coding: utf-8

"""Program to instantiate all the VIS tables in a database.

If a table already exists, no CREATE TABLE statement will be issued to
the database.
"""

# Copyright 2022-2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
import logging

from geoalchemy2 import load_spatialite  # type: ignore
from sqlalchemy import create_engine, insert, inspect, select
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

from vipersci import util

from vipersci.vis.db.image_records import ImageRecord
from vipersci.vis.db.image_requests import ImageRequest
from vipersci.vis.db.image_stats import ImageStats
from vipersci.vis.db.image_tags import ImageTag, taglist
from vipersci.vis.db.junc_image_pano import JuncImagePano
from vipersci.vis.db.junc_image_record_tags import JuncImageRecordTag
from vipersci.vis.db.junc_image_req_ldst import JuncImageRequestLDST
from vipersci.vis.db.ldst import LDST
from vipersci.vis.db.light_records import LightRecord
from vipersci.vis.db.pano_records import PanoRecord
from vipersci.vis.db.ptu_records import PanRecord, TiltRecord

# As new tables are defined, their Classes must be imported above, and
# then also added to this tuple:
tables = (
    ImageRecord,
    ImageRequest,
    ImageStats,
    ImageTag,
    JuncImagePano,
    JuncImageRecordTag,
    JuncImageRequestLDST,
    LDST,
    LightRecord,
    PanoRecord,
    PanRecord,
    TiltRecord,
)

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "-d",
        "--dburl",
        default="postgresql://postgres:NotTheDefault@localhost/visdb",
        help="Something like  %(default)s",
    )
    parser.add_argument(
        "-l", "--ldst", help="Path to semi-colon CSV file with LDST info."
    )
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    engine = create_engine(args.dburl)
    if args.dburl.startswith("sqlite://"):
        # This required because we have spatialite tables in the db:
        listen(engine, "connect", load_spatialite)

    # Create tables
    for t in tables:
        logger.info(f"Attempting to create {t.__tablename__}.")
        t.metadata.create_all(engine)

    with Session(engine) as session:
        # Establish image_tags
        scalars = session.scalars(select(ImageTag))
        results = scalars.all()
        if len(results) == 0:
            session.execute(insert(ImageTag), [{"name": x} for x in taglist])
            session.commit()
        elif len(results) == len(taglist):
            for i, row in enumerate(results):
                if row.name != taglist[i]:
                    raise ValueError(
                        f"Row {i} in the database has id {row.id} and tag {row.name} "
                        f"but should have {taglist[i]} from {taglist}."
                    )
        else:
            raise ValueError(
                f"The {ImageTag.__tablename__} table already contains the following "
                f"{len(results)} entries: {[r.name for r in results]}, but should "
                f"contain these {len(taglist)} entries: {taglist}"
            )

        # Set up LDST table
        if args.ldst is not None:
            with open(args.ldst, newline="") as csvfile:
                ldst_rows = list()
                reader = csv.reader(csvfile, delimiter=";")
                next(reader)  # Skip first line.
                next(reader)  # Skip second line.
                for row in reader:
                    ldst_rows.append(row)

            scalars = session.scalars(select(LDST))
            results = scalars.all()
            if len(results) == 0:
                session.execute(
                    insert(LDST), [{"id": x, "description": y} for (x, y) in ldst_rows]
                )
                session.commit()
            elif len(results) == len(ldst_rows):
                for i, row in enumerate(results):
                    if row.id != ldst_rows[i][0] or row.description != ldst_rows[i][1]:
                        raise ValueError(
                            f"Row {i} in the database has these values: {row} "
                            f"but should have {ldst_rows[i]}"
                        )
            else:
                raise ValueError(
                    f"The {LDST.__tablename__} table already contains the following "
                    f"{len(results)} entries: {results}, but should contain "
                    f"{len(ldst_rows)} entries."
                )

    # Check table names exists via inspect
    ins = inspect(engine)
    print("The following tables are now present in the database:")
    for t in ins.get_table_names():
        print(f"  {t}")



#!/usr/bin/env python
"""This module has tests for dotmap.generate_dotmap"""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.
import numpy as np
import rasterio
from vipersci.carto.dotmap import generate_dotmap


def test_single_point():
    """
    Test creating a raster with 1 data point
    """
    x_arr = [0]
    y_arr = [0]
    value_arr = [1]
    transform, raster = generate_dotmap(
        x_arr, y_arr, value_arr, radius=1, ground_sample_distance=1, padding=0
    )

    assert (-1, -1, 1, 1) == rasterio.windows.bounds(
        rasterio.windows.Window(0, 0, raster.shape[1], raster.shape[0]), transform
    )
    assert np.array_equal(raster, [[1, 1], [1, 1]])


def test_single_point_padded():
    """
    Test creating a raster with 1 data point and padding
    """
    x_arr = [0]
    y_arr = [0]
    value_arr = [1]
    transform, raster = generate_dotmap(
        x_arr, y_arr, value_arr, radius=1, ground_sample_distance=1, padding=1
    )
    assert (-2, -2, 2, 2) == rasterio.windows.bounds(
        rasterio.windows.Window(0, 0, raster.shape[1], raster.shape[0]), transform
    )
    assert np.array_equal(
        raster, [[-1, -1, -1, -1], [-1, 1, 1, -1], [-1, 1, 1, -1], [-1, -1, -1, -1]]
    )


def test_2_overlapping_points():
    """
    Test creating a raster with 2 overlapping data points
    """
    x_arr = [0, 1]
    y_arr = [0, 1]
    value_arr = [1, 2]
    transform, raster = generate_dotmap(
        x_arr, y_arr, value_arr, radius=1, ground_sample_distance=1, padding=0
    )

    assert (-1, -1, 2, 2) == rasterio.windows.bounds(
        rasterio.windows.Window(0, 0, raster.shape[1], raster.shape[0]), transform
    )
    assert np.array_equal(raster, [[-1, 2, 2], [1, 2, 2], [1, 1, -1]])


def test_single_detailed_point():
    """
    Test creating a raster with 1 data point and high raster resolution
    """
    x_arr = [0]
    y_arr = [0]
    value_arr = [1]
    transform, raster = generate_dotmap(
        x_arr, y_arr, value_arr, radius=1, ground_sample_distance=0.25, padding=0
    )

    assert (-1, -1, 1, 1) == rasterio.windows.bounds(
        rasterio.windows.Window(0, 0, raster.shape[1], raster.shape[0]), transform
    )
    assert np.array_equal(
        raster,
        [
            [-1] * 2 + [1] * 4 + [-1] * 2,
            [-1] + [1] * 6 + [-1],
            [1] * 8,
            [1] * 8,
            [1] * 8,
            [1] * 8,
            [-1] + [1] * 6 + [-1],
            [-1] * 2 + [1] * 4 + [-1] * 2,
        ],
    )


def test_horizontal_line():
    x_arr = [0, 1, 2, 3, 4]
    y_arr = [0] * len(x_arr)
    value_arr = [1] * len(x_arr)
    transform, raster = generate_dotmap(
        x_arr, y_arr, value_arr, radius=1, ground_sample_distance=0.5, padding=0
    )

    assert (-1, -1, 5, 1) == rasterio.windows.bounds(
        rasterio.windows.Window(0, 0, raster.shape[1], raster.shape[0]), transform
    )

    assert np.array_equal(
        raster,
        [
            [-1] + [1] * 10 + [-1],
            [1] * 12,
            [1] * 12,
            [-1] + [1] * 10 + [-1],
        ],
    )



# -*- coding: utf-8 -*-
"""
Takes a traverse of positions, a Burial Depth map, and a WEH map, and
simulates (with Poisson noise) what the Detector 1 and Detector 2 counts
from NSS might be at each location.

If no traverse is given, and both the Burial Depth map and the WEH map
have the same map characteristics, then instead of writing out a CSV file
that would simulate a traverse, this program will create a Detector1 map
and a Detector2 map that are the maps of the ideal detector values at
each location.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import csv
import logging
from pathlib import Path

import numpy as np
import rasterio

from vipersci import nss
from vipersci.carto.nss_modeler import write_tif

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    # Sooo many inputs each time.  Kind of a mess.
    # Should definitely set up @file config reading.
    parser.add_argument(
        "-b",
        "--burial_depth",
        type=Path,
        required=True,
        help="A GeoTIFF map whose pixels are the burial depth of the water " "layer. ",
    )
    parser.add_argument(
        "--det1",
        type=Path,
        required=True,
        help="A CSV file provided by the NSS team with the Detector 1 "
        "two-layer model.",
    )
    parser.add_argument(
        "--det2",
        type=Path,
        required=True,
        help="A CSV file provided by the NSS team with the Detector 2 "
        "two-layer model.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        required=True,
        help="The output CSV file or if -t is not given, this will be used "
        "as the prefix for the two output maps, which will end in "
        "_d1.tif and _d2.tif .",
    )
    parser.add_argument("-t", "--traverse", type=Path, help="Input CSV file.")
    parser.add_argument(
        "-w",
        "--weh",
        type=Path,
        required=True,
        help="A GeoTIFF map whose pixels are the water-equivalent-hydrogen "
        "percentage.",
    )

    return parser


class LocationSimulator:
    def __init__(
        self,
        bd_map: Path,
        weh_map: Path,
        det1: Path,
        det2: Path,
        bounds_error=True,
        fill_value=np.nan,
        rng=np.random.default_rng(),
    ):
        """
        :param bd_map: The path to the Burial Depth map.
        :type bd_map: Path
        :param weh_map: The path to the Water Equivalent Hydrogen map.
        :type weh_map: Path

        The other arguments (*det1*, *det2*, *bounds_error*, fill_value*,
        and *rng*) are passed to the nss.DataSimulator constructor,
        please see its documentation for more information.
        """
        self.bd_aff, self.bd_arr = self._init_map(bd_map)
        self.weh_aff, self.weh_arr = self._init_map(weh_map)

        self.ds = nss.DataSimulator(
            det1, det2, bounds_error=bounds_error, fill_value=fill_value, rng=rng
        )
        return

    @staticmethod
    def _init_map(raster: Path):
        # Helper function for __init__()
        data = rasterio.open(raster)
        return data.transform, data.read(1)

    def __call__(self, xycoords: np.ndarray, poisson=False):
        """
        Returns simulated detector 1 and detector 2 values at the given
        location(s).

        :param xycoords:  A sequence or np.array object of shape (2, ...)
        indicating the projected x, y coordinate to simulated the data for.
        :param poisson: A boolean (default False) indicating whether Poisson
        noise should be added to the returned simulated data.
        :returns: A two-tuple of values or np.arrays.  If xycoords contained
        only two elements, a two-tuple is returned of the detector 1 and
        detector 2 values at that location.  If xycoords contains more than
        one  location, then the returned two-tuple will be a numpy array
        of detector 1 values, and a numpy array of detector 2 values.
        """
        bd_rows, bd_cols = rasterio.transform.rowcol(
            self.bd_aff, xycoords[0], xycoords[1]
        )
        weh_rows, weh_cols = rasterio.transform.rowcol(
            self.weh_aff, xycoords[0], xycoords[1]
        )

        bd_vals = self.bd_arr[bd_rows, bd_cols]
        weh_vals = self.weh_arr[weh_rows, weh_cols]

        return self.ds(bd_vals, weh_vals, poisson=poisson)


def main():
    args = arg_parser().parse_args()

    if args.traverse is None:
        # Make ideal detector maps
        bd_data = rasterio.open(args.burial_depth)
        weh_data = rasterio.open(args.weh)

        for prop in ("width", "height", "transform"):
            if getattr(bd_data, prop) != getattr(weh_data, prop):
                raise ValueError(
                    f"The {prop} value of the Burial Depth and WEH map are "
                    "different.  They must be the same to output detector maps."
                )

        ds = nss.DataSimulator(
            args.det1, args.det2, bounds_error=False, fill_value=None
        )

        d1, d2 = ds(bd_data.read(1).flatten(), weh_data.read(1).flatten())

        d1_arr = d1.reshape(bd_data.shape)
        d2_arr = d2.reshape(bd_data.shape)

        kwds = bd_data.profile
        # for Detector 1
        kwds["nodata"] = 31.0
        write_tif(args.output, "_d1.tif", d1_arr, kwds)
        # for Detector 2
        kwds["nodata"] = 60.0
        write_tif(args.output, "_d2.tif", d2_arr, kwds)
        return

    else:
        # Make an output traverse file.
        simulator = LocationSimulator(
            args.burial_depth,
            args.weh,
            args.det1,
            args.det2,
            bounds_error=False,
            fill_value=None,
        )

        with open(args.traverse, newline="") as csvin, open(
            args.output, "w", newline=""
        ) as csvout:
            reader = csv.DictReader(csvin)  # expecting at least x and y cols.
            fieldnames = reader.fieldnames + ["det1", "det2", "det1pois", "det2pois"]
            writer = csv.DictWriter(csvout, fieldnames=fieldnames)
            writer.writeheader()

            # Would it be faster to read in all the points, feed them all to the
            # simulator, and then write out points?  Maybe, iterating for now:
            for row in reader:
                coords = np.array((row["x"], row["y"]), dtype=float)
                # print(coords)
                # print(coords.shape)
                row["det1"], row["det2"] = simulator(coords)
                row["det1pois"], row["det2pois"] = simulator(coords, poisson=True)

                writer.writerow(row)

    return



#!/usr/bin/env python
"""This module has tests for heatmap.area_bin"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import unittest

import numpy as np
from rasterio import transform, windows

from vipersci.carto import heatmap as hm


class TestHeatmapAreaBin(unittest.TestCase):
    def test_area_bin_3x3(self):
        x_arr = [0.5, 1.5, 2.5]
        y_arr = [0.5, 1.5, 2.5]
        x, y = np.meshgrid(x_arr, y_arr)
        values = np.ones(len(x_arr) * len(y_arr))
        averages, counts = hm.area_bin(
            values,
            x.ravel(),
            y.ravel(),
            transform.from_bounds(0, 0, 3, 3, 3, 3),
            windows.Window(0, 0, 3, 3),
        )

        self.assertTrue(np.array_equal(averages, np.ones((3, 3))))
        self.assertTrue(np.array_equal(counts, np.ones((3, 3))))
        # TODO: proper frequency tests



"""Produces various statistics from images.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
from textwrap import dedent
from typing import Union
from pathlib import Path

import numpy as np
import numpy.typing as npt
from rasterio.windows import Window
from scipy.ndimage import generic_filter
from skimage.morphology import disk
from skimage.io import imread
from skimage import measure

from vipersci import util

logger = logging.getLogger(__name__)

ImageType = Union[npt.NDArray[np.uint16], npt.NDArray[np.uint8]]

UNDEREXPOSED_THRESHOLD = 4095 * 0.2
OVEREXPOSED_THRESHOLD = 4095 * 0.8


class CentralMask:
    def __init__(self, image):
        self.array = np.zeros_like(image, dtype=int)
        quarter_width = int(image.shape[0] / 4)
        self.disk = disk(quarter_width)
        self.window = Window(quarter_width, quarter_width, *self.disk.shape)
        self.array[self.window.toslices()] += self.disk


def arg_parser():
    parser = argparse.ArgumentParser(
        description=__doc__, parents=[util.parent_parser()]
    )
    parser.add_argument(
        "image",
        type=Path,
        help="Path to image file which will be read and statistics computed.",
    )
    return parser


def main():
    args = arg_parser().parse_args()
    util.set_logger(args.verbose)

    image = imread(str(args.image))

    print(pprint(image))

    print(f"aggregate_noise: {aggregate_noise(image)} DN")

    return


def aggregate_noise(image):
    # This seems to yield a high value, and is slooooooow
    logger.info("aggregate_noise start")
    logger.info("Start std filtering.")
    std_filtered = generic_filter(image, np.std, size=(3, 3))
    logger.info("Start mean filtering.")
    mean_filtered = generic_filter(image, np.mean, size=(3, 3))
    logger.info("Dividing.")
    return np.nansum(np.divide(std_filtered, mean_filtered, where=mean_filtered != 0))


def compute(
    image: ImageType,
    overexposed_thresh=OVEREXPOSED_THRESHOLD,
    underexposed_thresh=UNDEREXPOSED_THRESHOLD,
) -> dict:
    mask_indices = np.logical_and(
        np.logical_and(
            image > underexposed_thresh,
            image < overexposed_thresh,
        ),
        CentralMask(image).array,
    )
    d = {
        "blur": measure.blur_effect(image),
        "mean": np.mean(image),
        "std": np.std(image),
        "over_exposed": (image > overexposed_thresh).sum(),
        "under_exposed": (image < underexposed_thresh).sum(),
        "central_clear": np.count_nonzero(mask_indices),
        # "aggregate_noise": aggregate_noise(image)
        # number of dead pixels
    }

    return d


def pprint(
    image: ImageType,
    overexposed_thresh=OVEREXPOSED_THRESHOLD,
    underexposed_thresh=UNDEREXPOSED_THRESHOLD,
) -> str:
    d = compute(image, overexposed_thresh, underexposed_thresh)
    disk_size = CentralMask(image).disk.sum()

    s = dedent(
        f"""\
        blur: {d['blur']} (0 for no blur, 1 for maximal blur)
        mean: {d['mean']}
        std: {d['std']}
        over-exposed (>{overexposed_thresh}): {d['over_exposed']} pixels,\
         {100 * d['over_exposed'] / image.size} %
        under-exposed (<{underexposed_thresh}): {d['under_exposed']} pixels,\
         {100 * d['under_exposed'] / image.size} %
        central_clear: {d['central_clear']} pixels, \
         {100 * d['central_clear'] / disk_size} %
        """
        # aggregate_noise: {d['aggregate_noise']} DN
        # """
    )
    return s



#!/usr/bin/env python
"""This module has tests for the vis.db.create_vis_dbs functions."""

# Copyright 2023, vipersci developers.
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

from argparse import ArgumentParser
import unittest
from unittest.mock import patch

from geoalchemy2 import load_spatialite
from sqlalchemy import create_engine, insert
from sqlalchemy.event import listen
from sqlalchemy.orm import Session

from vipersci.vis.db import Base
from vipersci.vis.db import create_vis_dbs as cvd
from vipersci.vis.db.image_tags import ImageTag, taglist


class TestParser(unittest.TestCase):
    def test_arg_parser(self):
        p = cvd.arg_parser()
        self.assertIsInstance(p, ArgumentParser)
        # self.assertRaises(SystemExit, p.parse_args)
        d = vars(p.parse_args([]))
        self.assertIn("dburl", d)


class TestDatabase(unittest.TestCase):
    def setUp(self) -> None:
        self.engine = create_engine("sqlite:///:memory:")
        listen(self.engine, "connect", load_spatialite)
        self.session = Session(self.engine)

    def tearDown(self):
        Base.metadata.drop_all(self.engine)

    def test_main(self):
        pa_ret_val = cvd.arg_parser().parse_args(["-d", "foo"])
        with patch("vipersci.vis.db.create_vis_dbs.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val
            with patch(
                "vipersci.vis.db.create_vis_dbs.create_engine", return_value=self.engine
            ):
                cvd.main()

    def test_partial_taglist(self):
        Base.metadata.create_all(self.engine)
        self.session.execute(insert(ImageTag), {"name": taglist[0]})
        self.session.commit()

        pa_ret_val = cvd.arg_parser().parse_args(["-d", "foo"])
        with patch("vipersci.vis.db.create_vis_dbs.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val
            with patch(
                "vipersci.vis.db.create_vis_dbs.create_engine", return_value=self.engine
            ):
                self.assertRaises(ValueError, cvd.main)

    def test_full_taglist(self):
        Base.metadata.create_all(self.engine)
        bad_taglist = taglist.copy()
        bad_taglist[0] = "Not a valid tag."
        self.session.execute(insert(ImageTag), [{"name": x} for x in bad_taglist])
        self.session.commit()

        pa_ret_val = cvd.arg_parser().parse_args(["-d", "foo"])
        with patch("vipersci.vis.db.create_vis_dbs.arg_parser") as parser:
            parser.return_value.parse_args.return_value = pa_ret_val
            with patch(
                "vipersci.vis.db.create_vis_dbs.create_engine", return_value=self.engine
            ):
                self.assertRaises(ValueError, cvd.main)



#!/usr/bin/env python
# coding: utf-8

"""Provides some generic PDS structures."""

# Copyright 2023, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import enum


class Purpose(enum.Enum):
    # These definitions are taken from the allowable values for
    # Product_Observational/Observation_Area/Primary_Result_Summary/purpose
    # in the PDS4 Data Dictionary.
    CALIBRATION = "Data collected to determine the relationship between measurement "
    "values and physical units."
    CHECKOUT = "Data collected during operational tests."
    ENGINEERING = "Data collected about support systems and structures, which are "
    "ancillary to the primary measurements."
    NAVIGATION = "Data collected to support navigation."
    OBSERVATION_GEOMETRY = "Data used to compute instrument observation geometry, "
    "such as SPICE kernels."
    SCIENCE = "Data collected primarily to answer questions about the targets of "
    "the investigation."
    SUPPORTING_OBSERVATION = "A science observation that was acquired to provide "
    "support for another science observation (e.g., a context image for a very "
    "high resolution observation, or an image intended to support an observation "
    "by a spectral imager)."



# -*- coding: utf-8 -*-
"""
Takes a pair of maps where one map has values for Detector 1 and the other
map has values for Detector 2, and produces Burial Depth and WEH% maps.
"""

# Copyright 2022, United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
#
# Reuse is permitted under the terms of the license.
# The AUTHORS file and the LICENSE file are at the
# top level of this library.

import argparse
import logging
from pathlib import Path

import numpy as np
import rasterio

from vipersci import nss

logger = logging.getLogger(__name__)


def arg_parser():
    parser = argparse.ArgumentParser(description=__doc__)
    # Sooo many inputs each time.  Kind of a mess.
    # Should definitely set up @file config reading.
    parser.add_argument(
        "--bd_mod",
        type=Path,
        required=True,
        help="A CSV file provided by the NSS team with the Burial Depth model.",
    )
    parser.add_argument("--det1", type=Path, required=True, help="A GeoTIFF file.")
    parser.add_argument("--det2", type=Path, required=True, help="A GeoTIFF file.")
    parser.add_argument(
        "-o",
        "--output",
        default="nss_model_",
        type=Path,
        help="The output prefix.  Will create three TIFF files that end in "
        "'bd.tif', 'weh.tif', and 'uweh.tif' and start with this. "
        "Default: %(default)s",
    )
    parser.add_argument(
        "--weh_mod",
        type=Path,
        required=True,
        help="A CSV file provided by the NSS team with the WEH model.",
    )

    return parser


def main():
    args = arg_parser().parse_args()

    det1_data = rasterio.open(args.det1)
    det1 = det1_data.read(1, masked=True)
    det2_data = rasterio.open(args.det2)
    det2 = det2_data.read(1, masked=True)

    nodata_val = -1
    modeler = nss.DataModeler(args.bd_mod, args.weh_mod, nodata_val)
    bd_arr, weh_arr, uweh_arr = modeler(det1, det2)

    kwds = det1_data.profile
    kwds["nodata"] = nodata_val
    kwds["dtype"] = np.double

    write_tif(args.output, "bd.tif", bd_arr, kwds)
    write_tif(args.output, "weh.tif", weh_arr, kwds)
    write_tif(args.output, "uweh.tif", uweh_arr, kwds)

    return


def write_tif(path: Path, ending: str, arr: np.typing.ArrayLike, kwds: dict):
    p = path.with_name(path.name + ending)
    with rasterio.open(p, "w", **kwds) as dst_dataset:
        dst_dataset.write(arr, 1)
    return p
